/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 7 $ */
/* $Date: 9/28/06 5:41p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.17.0-GSC-DN/driver/dma.c $
 *
 * 7     9/28/06 5:41p Dnewbold
 * Updated to perform DMA with 8-bit, 16-bit and 32-bit data.
 *
 * 6     5/30/06 1:16p Dnewbold
 * Modified the DMA procedure to better clean up afterwards.
 *
 * 5     2/27/06 10:30a Dnewbold
 * Streamlined DMA operation.
 *
 * 4     2/22/06 10:20a Dnewbold
 * Added DIPHASE2 support.
 *
 * 3     2/01/06 2:54p Dnewbold
 * Fixed timeout bug. Improved cleanup code.
 *
 * 2     9/01/05 1:26p Dnewbold
 * Added an event condition variable for the 2.6 kernel and fixed a
 * DMA/IRQ bug.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/******************************************************************************
*
*	Function:	_dma_channel_get
*
*	Purpose:
*
*		Get a PLX DMA channel for the specified operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		ability	The required read/write ability.
*
*	Returned:
*
*		< 0	The request failed.
*		0	Use channel zero.
*		1	Use channel one.
*
******************************************************************************/

static int _dma_channel_get(dev_data_t* dev, unsigned int ability)
{
	int	status;

	for (;;)	/* Use a loop for convenience. */
	{
		status	= down_interruptible(&dev->dma.sem);

		if (status)
		{
			status	= -1;
			break;
		}

		/*
		*	Check channel one first as it has write ability only
		*	and may be absent.
		*/

		if (	(dev->dma.channel[1].in_use == FALSE) &&
			(dev->dma.channel[1].flags & ability))
		{
			dev->dma.channel[1].in_use	= TRUE;
			status	= 1;
			up(&dev->dma.sem);
			break;
		}

		/*
		*	Check channel 0 second. We wouldn't want to give
		*	channel 0 for a write operation, as a read request
		*	would never get a DMA channel since it could never get
		*	channel 1 due to its limitations.
		*/

		if (	(dev->dma.channel[0].in_use == FALSE) &&
			(dev->dma.channel[0].flags & ability))
		{
			dev->dma.channel[0].in_use	= TRUE;
			status	= 0;
			up(&dev->dma.sem);
			break;
		}

		status	= -1;
		up(&dev->dma.sem);
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	dma_perform
*
*	Purpose:
*
*		Perform a DMA transfer per the given arguments.
*
*		NOTES:
*
*		With interrupts it is possible for the requested interrupt to
*		occure before the current task goes to sleep. This is possible,
*		for example, because the DMA transfer could complete before
*		the task gets to sleep. So, to make sure things work in these
*		cases, we take manual steps to insure that the current task
*		can be awoken by the interrupt before the task actually sleeps
*		and before the interrupt is enabled.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		queue		The wait queue to use for timeouts.
*
*		cond		The condition variable for the 2.6 kernel
*				event handling feature.
*
*		jit_end		The amount of time left in our timeout period.
*				This is in jiffies.
*
*		ability		The required read/write ability.
*
*		direction	The transfer direction bit for the DPR.
*
*		type		The transfer type bit for the MODE.
*
*		buff		The data transfer buffer.
*
*		samples		The number of samples to transfer.
*
*	Returned:
*
*		>= 0		The number of samples transferred.
*		< 0		There was an error.
*
******************************************************************************/

ssize_t dma_perform(	dev_data_t*		dev,
			WAIT_QUEUE_HEAD_T*	queue,
			int*			cond,
			unsigned long		jit_end,
			unsigned int		ability,
			unsigned int		direction,
			unsigned int		type,
			int			sample_bits,
			void*			buff,
			ssize_t			samples)
{
	u8			csr;
	dev_dma_t*		dma;
	int			index;
	ssize_t			qty;
	u32			size;
	long			timeout;
	u32			ul;
	WAIT_QUEUE_ENTRY_T	wait;

	for (;;)	/* Use a loop for convenience. */
	{
		index	= _dma_channel_get(dev, ability);

		if (index < 0)
		{
			/* We couldn't get a DMA channel. */
			qty	= 0;
			break;
		}

		dma	= &dev->dma.channel[index];

		/* Enable the interrupt. */
		irq_intcsr_mod(dev, dma->irq_enable, dma->irq_enable);

		if (	(dev->dma.bcr_dir_bit) &&
			(index == 0) &&
			(type == DMA_MODE_DEMAND_MODE_DMA))
		{
			/*
			*	Set the direction bit for DMA channel 0 when
			*	Demand Mode DMA is in use. We do this as the
			*	application may have enabled both the transmit
			*	and the receive options enabled and Demand
			*	Mode DMA can only go one way at a time.
			*/

			ul	= readl(dev->vaddr.gsc_bcr_32);

			if (direction == DMA_DPR_HOST_TO_BOARD)
			{
				ul |= HPDI32_GSC_BCR_TX_DMDMA0;
				writel(ul, dev->vaddr.gsc_bcr_32);
			}
			else
			{
				ul &= ~HPDI32_GSC_BCR_TX_DMDMA0;
				writel(ul, dev->vaddr.gsc_bcr_32);
			}
		}

		/* DMAMODEx */

		if (sample_bits == HPDI32_IO_SAMPLE_BITS_32)
			size	= DMA_MODE_SIZE_32_BITS;
		else if (sample_bits == HPDI32_IO_SAMPLE_BITS_16)
			size	= DMA_MODE_SIZE_16_BITS;
		else
			size	= DMA_MODE_SIZE_8_BITS;

		ul	= type
			| size
			| DMA_MODE_INPUT_ENABLE
			| DMA_MODE_BURSTING_LOCAL
			| DMA_MODE_INTERRUPT_WHEN_DONE
			| DMA_MODE_LOCAL_ADRESS_CONSTANT
			| DMA_MODE_PCI_INTERRUPT_ENABLE;
		writel(ul, dma->vaddr.mode_32);

		/* DMAPADRx */
		ul	= virt_to_bus(buff);
		writel(ul, dma->vaddr.padr_32);

		/* DMALADRx */
		ul	= HPDI32_REG_OFFSET(HPDI32_GSC_FDR);
		writel(ul, dma->vaddr.ladr_32);

		/* DMASIZx */
		ul	= samples * (sample_bits / 8);
		writel(ul, dma->vaddr.siz_32);

		/* DMADPRx */
		ul	= direction
			| DMA_DPR_END_OF_CHAIN
			| DMA_DPR_TERMINAL_COUNT_IRQ;
		writel(ul, dma->vaddr.dpr_32);

		/* DMAARB */
		writel(0, dev->vaddr.plx_dmaarb_32);

		/* DMATR */
		writel(0, dev->vaddr.plx_dmathr_32);

		/*
		*	Do the magic!
		*/

		dma->queue	= queue;
		dma->condition	= cond;
		memset(&wait, 0, sizeof(wait));
		WAIT_QUEUE_ENTRY_INIT(&wait, current);	/* NOTES */
		SET_CURRENT_STATE(TASK_INTERRUPTIBLE);	/* NOTES */
		add_wait_queue(queue, &wait);		/* NOTES */

		writeb(0, dma->vaddr.csr_8);		/* Disabe */
		ul	= DMA_CSR_ENABLE;
		writeb(ul, dma->vaddr.csr_8);		/* Enable */
		ul	= DMA_CSR_ENABLE | DMA_CSR_START;
		writeb(ul, dma->vaddr.csr_8);		/* Start */

		timeout = jit_end ? (jit_end - jiffies) : HZ;
		timeout	= (timeout <= 0) ? 1 : timeout;
		EVENT_WAIT_IRQ_TO(dma->queue, dma->condition[0], timeout);
		remove_wait_queue(queue, &wait);	/* NOTES */
		SET_CURRENT_STATE(TASK_RUNNING);	/* NOTES */

		/*
		*	Clean up.
		*/

		// Check the results.
		csr	= readb(dma->vaddr.csr_8);

		/* Disable the DONE interrupt. */
		ul	= readl(dma->vaddr.mode_32);
		ul	&= ~DMA_MODE_INTERRUPT_WHEN_DONE;
		writel(ul, dma->vaddr.mode_32);


		if ((csr & DMA_CSR_DONE) == 0)
		{
			/*
			*	If it isn't done then the operation timed out.
			*	Unfortunately the PLX PCI9080 doesn't tell us
			*	how many bytes were transferred. So, instead
			*	of reporting the number of transferred bytes,
			*	or zero, we return an error status.
			*/

			/* Force DMA termination. */
			writeb(DMA_CSR_ABORT, dma->vaddr.csr_8);
			udelay(10);		/* Wait for completion. */
			qty	= -ETIMEDOUT;
		}
		else
		{
			qty	= samples;
		}

		/* Disable the DMA interrupt. */
		irq_intcsr_mod(dev, 0, dma->irq_enable);

		/* Clear the DMA. */
		writeb(DMA_CSR_CLEAR, dma->vaddr.csr_8);

		/* Let someone else use the DMA channel. */
		dma->queue	= NULL;
		dma->condition	= NULL;
		dma->in_use	= FALSE;
		break;
	}

	return(qty);
}



