/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 4 $ */
/* $Date: 9/28/06 5:43p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.17.0-GSC-DN/driver/io.c $
 *
 * 4     9/28/06 5:43p Dnewbold
 * Added code to support the Data Size and PIO Threshold parameters.
 *
 * 3     2/01/06 2:56p Dnewbold
 * Updated the I/O buffer allocation mechanism to permit larger buffers.
 *
 * 2     9/01/05 1:27p Dnewbold
 * Added an event condition variable for the 2.6 kernel.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/******************************************************************************
*
*	Function:	_order_to_size
*
*	Purpose:
*
*		Convert an order value to its corresponding size in page sized
*		units.
*
*	Arguments:
*
*		order	The page size order to convert.
*
*	Returned:
*
*		>= 0	The order corresponding to the given size.
*
******************************************************************************/

u32 _order_to_size(int order)
{
	u32	size	= PAGE_SIZE;

	// Limit sizes to 1G.

	for (; order > 0; order--)
	{
		if (size >= 0x40000000)
			break;

		size	*= 2;
	}

	return(size);
}



/******************************************************************************
*
*	Function:	_size_to_order
*
*	Purpose:
*
*		Convert a size value to a corresponding page size order value.
*
*	Arguments:
*
*		size	The desired size in bytes.
*
*	Returned:
*
*		>= 0	The order corresponding to the given size.
*
******************************************************************************/

int _size_to_order(u32 size)
{
	u32	bytes	= PAGE_SIZE;
	int	order	= 0;

	// Limit sizes to 1G.

	for (; bytes < 0x40000000;)
	{
		if (bytes >= size)
			break;

		order++;
		bytes	*= 2;
	}

	return(order);
}



/******************************************************************************
*
*	Function:	_pages_alloc
*
*	Purpose:
*
*		Allocate a block of memory that based on pages. The allocated
*		block is on a page boundary and the allocated size is rounded
*		upto a page boundary.
*
*	Arguments:
*
*		samples	The desired number of samples.
*
*		io	The structure through which the request data is
*			received and the allocation is returned.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _pages_alloc(u32 samples, dev_io_t* io)
{
	unsigned long	adrs;
	s32		bytes;
	u8		order;
	unsigned long	ul;

	bytes		= samples * 4;

	if (bytes > MEM_ALLOC_LIMIT)
		bytes	= MEM_ALLOC_LIMIT;

	order	= _size_to_order(bytes);
	bytes	= _order_to_size(order);

	for (;;)
	{
		adrs	= __get_free_pages(GFP_KERNEL | __GFP_NOWARN, order);

		if (adrs)
			break;

		if (order == 0)
			break;

		order--;
		bytes	/= 2;
	}

	if (adrs == 0)
	{
		io->adrs	= 0;
		io->buf		= NULL;
		io->bytes	= bytes;
		io->order	= 0;
		io->samples	= 0;
	}
	else
	{
		io->adrs	= adrs;
		io->buf		= (u32*) adrs;
		io->bytes	= bytes;
		io->order	= order;
		io->samples	= bytes / 4;

		/* Mark the pages as reserved so the MM will leave'm alone. */
		ul	= (unsigned long) io->buf;

		for (; bytes >= PAGE_SIZE; bytes -= PAGE_SIZE, ul += PAGE_SIZE)
			PAGE_RESERVE(ul);
	}
}



/******************************************************************************
*
*	Function:	_pages_free
*
*	Purpose:
*
*		Free memory allocated via _pages_alloc().
*
*	Arguments:
*
*		io	Parameters for the allocation to be freed.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _pages_free(dev_io_t* io)
{
	s32		bytes;
	unsigned long	ul;

	if (io->adrs)
	{
		/* Unreserve the pages before we return them to the MM. */
		bytes	= io->bytes;
		ul	= (unsigned long) io->buf;

		for (; bytes >= PAGE_SIZE; bytes -= PAGE_SIZE, ul += PAGE_SIZE)
			PAGE_UNRESERVE(ul);

		/* Free the memory. */
		free_pages(io->adrs, io->order);
		io->adrs	= 0;
		io->buf		= NULL;
		io->bytes	= 0;
		io->order	= 0;
		io->samples	= 0;
	}
}



/******************************************************************************
*
*	Function:	_io_mode_validate
*
*	Purpose:
*
*		Validate an I/O mode argument to the I/O Configure IOCTL
*		service.
*
*	Arguments:
*
*		dev	The current recorded value.
*
*		src	The value to be validated.
*
*		dest	The results are recorded here.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _io_mode_validate(s32 dev, s32 src, s32* dest)
{
	int	status;

	switch (src)
	{
		default:

			status	= -EINVAL;
			break;

		case HPDI32_IO_MODE_NO_CHANGE:

			status	= 0;
			dest[0]	= dev;
			break;

		case HPDI32_IO_MODE_DMA:
		case HPDI32_IO_MODE_DMDMA:
		case HPDI32_IO_MODE_PIO:

			status	= 0;
			dest[0]	= src;
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_sample_bits_validate
*
*	Purpose:
*
*		Validate an I/O sample_bits argument to the I/O Configure
*		IOCTL service.
*
*	Arguments:
*
*		dev	The current recorded value.
*
*		src	The value to be validated.
*
*		dest	The results are recorded here.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _io_sample_bits_validate(s32 dev, s32 src, s32* dest)
{
	int	status;

	switch (src)
	{
		default:

			status	= -EINVAL;
			break;

		case HPDI32_IO_SAMPLE_BITS_NO_CHANGE:

			status	= 0;
			dest[0]	= dev;
			break;

		case HPDI32_IO_SAMPLE_BITS_8:
		case HPDI32_IO_SAMPLE_BITS_16:
		case HPDI32_IO_SAMPLE_BITS_32:

			status	= 0;
			dest[0]	= src;
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_samples_validate
*
*	Purpose:
*
*		Validate a samples argument to the I/O Configure IOCTL service.
*
*	Arguments:
*
*		mmap	The current mmap() usage count.
*
*		dev	The current recorded value.
*
*		src	The value to be validated.
*
*		dest	The results are recorded here.
*
*		size	The size of the corresponding FIFO. This is zero if
*			the size is unknown.
*
*		io	The resulting allocation, if any, is recorded here.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _io_samples_validate(int		mmap,
				u32		dev,
				u32		src,
				u32*		dest,
				u32		size,
				dev_io_t*	io)
{
	int	status;

	if ((src == dev) || (src == HPDI32_IO_SAMPLES_NO_CHANGE))
	{
		/* No change is being mad. */
		status	= 0;
		dest[0]	= dev;
	}
	else if (src < HPDI32_IO_SAMPLES_MIN)
	{
		/* The requested size is too small. */
		status	= -EINVAL;
	}
	else if (src > HPDI32_IO_SAMPLES_MAX)
	{
		/* The requested size is too big. */
		status	= -EINVAL;
	}
	else if (mmap)
	{
		/* The buffer can't be changed while mmap() is in use. */
		status	= -EBUSY;
	}
	else
	{
		if ((size) && (src > size))
			dest[0]	= size;
		else
			dest[0]	= src;

		_pages_alloc(dest[0], io);

		if (io->adrs == 0)
		{
			status	= -ENOMEM;
		}
		else
		{
			status	= 0;
			dest[0]	= io->samples;
		}
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_pio_threshold_validate
*
*	Purpose:
*
*		Validate a PIO Threshold argument to the I/O Configure IOCTL
*		service.
*
*	Arguments:
*
*		dev	The current recorded value.
*
*		src	The value to be validated.
*
*		dest	The results are recorded here.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _io_pio_threshold_validate(s32 dev, s32 src, s32* dest)
{
	int	status	= 0;

	if (src == HPDI32_IO_PIO_THRESHOLD_NO_CHANGE)
		dest[0]	= dev;
	else if (src < HPDI32_IO_PIO_THRESHOLD_MIN)
		status	= -EINVAL;
	else
		dest[0]	= src;

	return(status);
}



/******************************************************************************
*
*	Function:	_io_timeout_validate
*
*	Purpose:
*
*		Validate a timeout argument to the I/O Configure IOCTL service.
*
*	Arguments:
*
*		dev	The current recorded value.
*
*		src	The value to be validated.
*
*		dest	The results are recorded here.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _io_timeout_validate(s32 dev, s32 src, s32* dest)
{
	int	status	= 0;

	if (src == HPDI32_IO_TIMEOUT_NO_CHANGE)
		dest[0]	= dev;
	else if (src < HPDI32_IO_TIMEOUT_MIN)
		status	= -EINVAL;
	else if (src > HPDI32_IO_TIMEOUT_MAX)
		status	= -EINVAL;
	else
		dest[0]	= src;

	return(status);
}



/******************************************************************************
*
*	Function:	_read_work
*
*	Purpose:
*
*		Implement the mode independent working portion of the read()
*		procedure. If a timeout is permitted and called for, we use
*		Rx FIFO status to wake the task when the corresponding status
*		has changed. If that interrupt is not available to use, then
*		we merely wait a short period as needed instead.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		buff		The samples read are placed here. This is NULL
*				when mmap() is in use.
*
*		samples		The number of samples requested. This is
*				positive.
*
*		jit_end		End at this point in time (in jiffies). This
*				is zero if we're to end as soon at no data is
*				available.
*
*		fn_available	This is the mode specific function that
*				determines the number of samples that may be
*				read.
*
*		fn_work		This is the function implementing the mode
*				specific reading functionality.
*
*		irq_index	The IRQ index to awaken on when we run out of
*				data.
*
*	Returned:
*
*		>= 0	This is the number of samples read.
*
******************************************************************************/

static ssize_t _read_work(	dev_data_t*	dev,
				unsigned char*	buff,
				ssize_t		samples,
				unsigned long	jit_end,
				ssize_t		(*fn_available)(
						dev_data_t*	dev,
						ssize_t		samples),
				ssize_t		(*fn_work)(
						dev_data_t*	dev,
						u8*		buff,
						ssize_t		samples,
						unsigned long	jit_end),
				int		irq_index)
{
	ssize_t		bytes_to_user		= 0;
	int		bytes_per_sample	= dev->read.sample_bits / 8;
	unsigned long	bytes_from_board       	= 0;
	u8*		dest			= (u8*) dev->read.buf;
	ssize_t		qty;
	ssize_t		samples_from_board	= 0;
	int		test;
	unsigned long	ul;

	for (;;)
	{
		if (samples_from_board <= 0)
		{
			/* Transfer one block at a time. */

			if (bytes_from_board > 0)
			{
				if (buff)
				{
					/*
					*	Copy the data to user space. We
					*	don't do this copy when mmap()
					*	is in use.
					*/

					ul	= copy_to_user(
							buff,
							dev->read.buf,
							bytes_from_board);

					if (ul)
					{
						/* We couldn't get user data. */
						bytes_from_board	= 0;
						bytes_to_user	= -EFAULT;
						break;
					}

					buff	+= bytes_from_board;
				}

				bytes_to_user	+= bytes_from_board;
			}

			/* Get a full or partial block? */

			if (samples > dev->read.samples)
				samples_from_board	= dev->read.samples;
			else
				samples_from_board	= samples;

			bytes_from_board	= 0;
			dest			= (u8*) dev->read.buf;
		}

		qty	= (fn_available)(dev, samples_from_board);

		if (qty)
		{
			qty	= (fn_work)(dev, dest, qty, jit_end);

			if (qty < 0)
			{
				/* There was a problem. */
				bytes_from_board	= 0;
				bytes_to_user		= qty;
				break;
			}

			samples			-= qty;
			samples_from_board	-= qty;
			dest			+= qty * bytes_per_sample;
			bytes_from_board	+= qty * bytes_per_sample;

			if (samples == 0)
			{
				/* The request has been fulfilled. */
				break;
			}
			else if ((jit_end) && (jiffies > jit_end))
			{
				/* We got some data then timed out. */
				break;
			}
			else if (jit_end)
			{
				/* We've got more time to try for more. */
				continue;
			}
			else if (qty)
			{
				/* We can't block bit we did get some data. */
				continue;
			}
			else
			{
				/* We can't block and we got no data. */
				break;
			}
		}
		else if (jit_end == 0)
		{
			/* No data is available and we can't block. */
			break;
		}
		else if ((jit_end) && (jiffies > jit_end))
		{
			/*
			*	No data is available and we can't wait any
			*	longer for more.
			*/

			break;
		}

		/* Wait a while for data to become available. */
		test	= irq_gsc_acquire(	dev,
						&dev->read.queue,
						&dev->read.condition,
						jit_end,
						irq_index,
						_INT_EL_LEVEL,
						_INT_HL_HIGH);

		if (test == IRQ_GSC_IRQ)
			continue;			/* Got the interrupt. */
		else if (test == IRQ_GSC_TIME_SIG)
			break;				/* Timeout or signal. */
		else
			udelay(10);			/* Didn't get IRQ. */
	}

	if ((bytes_from_board > 0) && (buff))
	{
		/*
		*	Copy the data to user space. We don't do this copy
		*	when mmap() is in use.
		*/

		ul	= copy_to_user(	buff,
					dev->read.buf,
					bytes_from_board);

		if (ul)
			bytes_to_user	= -EFAULT;
		else
			bytes_to_user	+= bytes_from_board;
	}
	else
	{
		bytes_to_user	+= bytes_from_board;
	}

	return(bytes_to_user);
}



/******************************************************************************
*
*	Function:	_write_work
*
*	Purpose:
*
*		Implement the mode independent portion of the write()
*		procedure. If a timeout is permitted and called for, we use
*		the FIFO status to wake the task when the appropriate amount
*		of space becomes available. If that interrupt is not available
*		to use, then we merely wait a short period as needed instead.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		buff		The samples read are placed here. This will be
*				NULL when mmap() is in use.
*
*		samples	T	The number of samples requested. This is
*				positive.
*
*		jit_end		End at this point in time (in jiffies). This
*				is zero if we're to end as soon at no data is
*				available.
*
*		fn_available	This is the mode specific function that
*				determines the number of samples that may be
*				read.
*
*		fn_work		This is the function implementing the mode
*				specific reading functionality.
*
*		irq_index	The IRQ index to awaken on when we run out of
*				data.
*
*	Returned:
*
*		>= 0	This is the number of samples written.
*
******************************************************************************/

static ssize_t _write_work(
		dev_data_t*		dev,
		const unsigned char*	buff,
		ssize_t			samples,
		unsigned long		jit_end,
		ssize_t			(*fn_available)(
					dev_data_t*	dev,
					ssize_t		samples),
		ssize_t			(*fn_work)(
					dev_data_t*		dev,
					const u8*		buff,
					ssize_t			samples,
					unsigned long		jit_end),
		int			irq_index)
{
	ssize_t		bytes_from_user		= 0;
	int		bytes_per_sample	= dev->write.sample_bits / 8;
	unsigned long	bytes_to_board;
	ssize_t		qty;
	ssize_t		samples_to_board	= 0;
	u8*		src			= (u8*) dev->write.buf;
	int		test;
	unsigned long	ul;

	for (;;)
	{
		if (samples_to_board <= 0)
		{
			/* Transfer one block at a time. */

			if (samples > dev->write.samples)
				samples_to_board	= dev->write.samples;
			else
				samples_to_board	= samples;

			bytes_to_board	= samples_to_board * bytes_per_sample;
			src		= (u8*) dev->write.buf;

			if (buff)
			{
				ul	= copy_from_user(	src,
								buff,
								bytes_to_board);
				buff	+= bytes_to_board;

				if (ul)
				{
					/* We couldn't get user data. */

					if (bytes_from_user == 0)
						bytes_from_user	= -EFAULT;

					break;
				}
			}
		}

		qty	= (fn_available)(dev, samples_to_board);

		if (qty)
		{
			qty	= (fn_work)(dev, src, qty, jit_end);

			if (qty < 0)
			{
				/* There was a problem. */
				bytes_from_user	= qty;
				break;
			}

			samples_to_board	-= qty;
			samples			-= qty;
			src			+= qty * bytes_per_sample;
			bytes_from_user		+= qty * bytes_per_sample;

			if (samples == 0)
			{
				/* The request has been fulfilled. */
				break;
			}
			else if ((jit_end) && (jiffies > jit_end))
			{
				/* Some data was written then we timed out. */
				break;
			}
			else if (jit_end)
			{
				/* We've got more time to try for more. */
				continue;
			}
			else if (qty)
			{
				/* We can't block but we did send some data. */
				continue;
			}
			else
			{
				/* We can't block and we sent no data. */
				break;
			}
		}
		else if (jit_end == 0)
		{
			/* No space is available and we can't block. */
			break;
		}
		else if ((jit_end) && (jiffies > jit_end))
		{
			/*
			*	No space is available and we can't wait any
			*	longer for more.
			*/

			break;
		}

		/* Wait a while for space to become available. */
		test	= irq_gsc_acquire(	dev,
						&dev->write.queue,
						&dev->write.condition,
						jit_end,
						irq_index,
						_INT_EL_LEVEL,
						_INT_HL_HIGH);

		if (test == IRQ_GSC_IRQ)
			continue;			/* Got the interrupt. */
		else if (test == IRQ_GSC_TIME_SIG)
			break;				/* Timeout or signal. */
		else
			udelay(10);			/* Didn't get IRQ. */
	}

	return(bytes_from_user);
}



/******************************************************************************
*
*	Function:	io_close_dev
*
*	Purpose:
*
*		Cleanup the I/O stuff for the device as it is being closed.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void io_close_dev(dev_data_t* dev)
{
	if (dev->read.buf)
		_pages_free(&dev->read);

	if (dev->write.buf)
		_pages_free(&dev->write);

	dev->read.buf	= NULL;
	dev->write.buf	= NULL;
}



/******************************************************************************
*
*	Function:	io_config_ioctl
*
*	Purpose:
*
*		Implement the I/O Configure IOCTL service. Here, we validate
*		the arguments then, if valid, apply them and report the
*		current settings.
*
*		NOTE: Use care in modifying the function as it is used by
*			the driver as well as by user applications.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int io_config_ioctl(dev_data_t* dev, hpdi32_io_config_t* arg)
{
	int			status;
	dev_io_t		rx;
	int			rx_sem	= -1;
	int			test;
	hpdi32_io_config_t	tmp;
	dev_io_t		tx;
	int			tx_sem	= -1;

	for (;;)	/* A convenience loop. */
	{
		memset(&rx, 0, sizeof(rx));
		memset(&tx, 0, sizeof(tx));

		/*
		*	Get the read() and write() semaphores.
		*/

		rx_sem	= down_interruptible(&dev->read.sem);
		tx_sem	= down_interruptible(&dev->write.sem);

		if ((rx_sem) || (tx_sem))
		{
			/* We got a signal rather than the semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		/*
		*	Validate the mode arguments.
		*/

		status	= _io_mode_validate(	dev->read.mode,
						arg->read.mode,
						&tmp.read.mode);
		test	= _io_mode_validate(	dev->write.mode,
						arg->write.mode,
						&tmp.write.mode);
		status	= status ? status : test;

		if (status)
			break;

		/*
		*	Validate the PIO Threshold arguments.
		*/

		status	= _io_pio_threshold_validate(
				dev->read.pio_threshold,
				arg->read.pio_threshold,
				&tmp.read.pio_threshold);
		test	= _io_pio_threshold_validate(
				dev->write.pio_threshold,
				arg->write.pio_threshold,
				&tmp.write.pio_threshold);
		status	= status ? status : test;

		if (status)
			break;

		/*
		*	Validate the timeout arguments.
		*/

		status	= _io_timeout_validate(	dev->read.timeout,
						arg->read.timeout,
						&tmp.read.timeout);
		test	= _io_timeout_validate(	dev->write.timeout,
						arg->write.timeout,
						&tmp.write.timeout);
		status	= status ? status : test;

		if (status)
			break;

		/*
		*	Validate the sample size arguments.
		*/

		status	= _io_sample_bits_validate(	dev->read.sample_bits,
							arg->read.sample_bits,
							&tmp.read.sample_bits);
		test	= _io_sample_bits_validate(	dev->write.sample_bits,
							arg->write.sample_bits,
							&tmp.write.sample_bits);
		status	= status ? status : test;

		if (status)
			break;

		/*
		*	Validate the samples arguments and possibly
		*	allocate buffers.
		*/

		status	= _io_samples_validate(	dev->mmap.read.usage,
						dev->read.samples,
						arg->read.samples,
						&tmp.read.samples,
						dev->cache.gsc_rfsr,
						&rx);

		if (status)
			break;

		status	= _io_samples_validate(	dev->mmap.write.usage,
						dev->write.samples,
						arg->write.samples,
						&tmp.write.samples,
						dev->cache.gsc_tfsr,
						&tx);

		if (status)
			break;

		/*
		*	Apply the results.
		*/

		if (rx.adrs)
		{
			if (dev->read.buf)
				_pages_free(&dev->read);

			dev->read.adrs		= rx.adrs;
			dev->read.buf		= rx.buf;
			dev->read.bytes		= rx.bytes;
			dev->read.order		= rx.order;
			dev->read.samples	= rx.samples;

			tmp.read.samples	= rx.samples;

			rx.adrs			= 0;
			rx.buf			= NULL;
			rx.bytes		= 0;
			rx.order		= 0;
			rx.samples		= 0;
		}

		if (tx.adrs)
		{
			if (dev->write.buf)
				_pages_free(&dev->write);

			dev->write.adrs		= tx.adrs;
			dev->write.buf		= tx.buf;
			dev->write.bytes	= tx.bytes;
			dev->write.order	= tx.order;
			dev->write.samples	= tx.samples;

			tmp.write.samples	= tx.samples;

			tx.adrs			= 0;
			tx.buf			= NULL;
			tx.bytes		= 0;
			tx.order		= 0;
			tx.samples		= 0;
		}

		dev->read.mode			= tmp.read.mode;
		dev->read.timeout		= tmp.read.timeout;
		dev->read.samples		= tmp.read.samples;
		dev->read.sample_bits		= tmp.read.sample_bits;
		dev->read.pio_threshold		= tmp.read.pio_threshold;
		dev->write.mode			= tmp.write.mode;
		dev->write.timeout		= tmp.write.timeout;
		dev->write.samples		= tmp.write.samples;
		dev->write.sample_bits		= tmp.write.sample_bits;
		dev->write.pio_threshold	= tmp.write.pio_threshold;

		arg->read.mode			= tmp.read.mode;
		arg->read.timeout		= tmp.read.timeout;
		arg->read.samples		= tmp.read.samples;
		arg->read.sample_bits		= tmp.read.sample_bits;
		arg->read.pio_threshold		= tmp.read.pio_threshold;
		arg->write.mode			= tmp.write.mode;
		arg->write.timeout		= tmp.write.timeout;
		arg->write.samples		= tmp.write.samples;
		arg->write.sample_bits		= tmp.write.sample_bits;
		arg->write.pio_threshold	= tmp.write.pio_threshold;
		mmap_info_compute(dev);
		break;
	}

	/* Clean up. */

	if (rx.adrs)
		_pages_free(&rx);

	if (rx_sem == 0)
		up(&dev->read.sem);

	if (tx.adrs)
		_pages_free(&tx);

	if (tx_sem == 0)
		up(&dev->write.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	read
*
*	Purpose:
*
*		Implement the read() procedure. Only one read thread is
*		permitted at a time.
*
*	Arguments:
*
*		filp	This is the file structure.
*
*		buff	The data requested goes here.
*
*		count	The number of bytes requested.
*
*		offp	The file position the user is accessing.
*
*	Returned:
*
*		0	All went well, though no data was transferred.
*		> 0	This is the number of bytes transferred.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t read(	struct file*	filp,
		char*		buff,
		size_t		count,
		loff_t*		offp)
{
	dev_data_t*	dev;
	unsigned long	jit_end;
	size_t		samples;
	ssize_t		status;
	int		test;
	unsigned long	ul;

	for (;;)	/* We'll use a loop for convenience. */
	{
		/*
		*	Validate some of the parameters.
		*/

		if (count <= 0)
		{
			/* No data transfer called for. */
			status	= 0;
			break;
		}

		if (buff)
		{
			ul	= access_ok(VERIFY_WRITE, buff, count);

			if (ul == 0)
			{
				/* We can't write to the user's memory. */
				status	= -EFAULT;
				break;
			}
		}

		dev	= (dev_data_t*) filp->private_data;

		if (dev == NULL)
		{
			/* The referenced device doesn't exist. */
			status	= -ENODEV;
			break;
		}

		/*
		*	Gain the neccessary access.
		*/

		test	= down_interruptible(&dev->sem);

		if (test)
		{
			/* We got a signal rather than the semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		test	= down_interruptible(&dev->read.sem);
		up(&dev->sem);

		if (test)
		{
			/* We got a signal rather than the READ semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		/*
		*	Verify against mmap() usage.
		*/

		if ((dev->mmap.read.usage) && (buff))
		{
			/* Must be NULL when mmap() is in use. */
			status	= -EINVAL;
			up(&dev->read.sem);
			break;
		}

		// Verify the count.

		if (dev->read.sample_bits == HPDI32_IO_SAMPLE_BITS_32)
		{
			samples	= count / 4;

			if (count % 4)
			{
				// The size must be a multiple of 4 bytes.
				status	= -EINVAL;
				up(&dev->read.sem);
				break;
			}
		}
		else if (dev->read.sample_bits == HPDI32_IO_SAMPLE_BITS_16)
		{
			samples	= count / 2;

			if (count % 2)
			{
				// The size must be a multiple of 2 bytes.
				status	= -EINVAL;
				up(&dev->read.sem);
				break;
			}
		}
		else
		{
			samples	= count;
		}

		/*
		*	Prepare for the operation.
		*/

		if ((buff == NULL) && (samples > dev->read.samples))
		{
			/* Limit reads to the size of the buffer. */
			samples	= dev->read.samples;
		}

		if ((filp->f_flags & O_NONBLOCK) || (dev->read.timeout <= 0))
			jit_end	= 0;
		else
			jit_end	= jiffies + dev->read.timeout * HZ;

		/*
		*	Perform the operation.
		*/

		if (	(samples <= dev->read.pio_threshold) ||
			(dev->read.mode == HPDI32_IO_MODE_PIO))
		{
			status	= _read_work(	dev,
						buff,
						samples,
						jit_end,
						pio_read_available,
						pio_read_work,
						_INT_INDEX_RX_FIFO_EMPTY);
		}
		else if (dev->read.mode == HPDI32_IO_MODE_DMA)
		{
			status	= _read_work(	dev,
						buff,
						samples,
						jit_end,
						dma_read_available,
						dma_read_work,
						_INT_INDEX_RX_FIFO_AE);
		}
		else if (dev->read.mode == HPDI32_IO_MODE_DMDMA)
		{
			status	= _read_work(	dev,
						buff,
						samples,
						jit_end,
						dmdma_read_available,
						dmdma_read_work,
						_INT_INDEX_RX_FIFO_AE);
		}
		else
		{
			status	= -EINVAL;
		}

		/*
		*	Clean up.
		*/

		up(&dev->read.sem);
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	write
*
*	Purpose:
*
*		Implement the write() procedure. Only one thread is permitted
*		at a time.
*
*	Arguments:
*
*		filp	This is the file structure.
*
*		buff	The data requested goes here.
*
*		count	The number of bytes requested.
*
*		offp	The file position the user is accessing.
*
*	Returned:
*
*		0	All went well, though no data was transferred.
*		> 0	This is the number of bytes transferred.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t write(	struct file*	filp,
		const char*	buff,
		size_t		count,
		loff_t*		offp)
{
	dev_data_t*	dev;
	unsigned long	jit_end;
	size_t		samples;
	ssize_t		status;
	int		test;
	unsigned long	ul;

	for (;;)	/* We'll use a loop for convenience. */
	{
		/*
		*	Validate some of the parameters.
		*/

		if (count <= 0)
		{
			/* No data transfer called for. */
			status	= 0;
			break;
		}

		if (buff)
		{
			ul	= access_ok(VERIFY_READ, buff, count);

			if (ul == 0)
			{
				/* We can't read from the user's memory. */
				status	= -EFAULT;
				break;
			}
		}

		dev	= (dev_data_t*) filp->private_data;

		if (dev == NULL)
		{
			/* The referenced device doesn't exist. */
			status	= -ENODEV;
			break;
		}

		/*
		*	Gain the neccessary access.
		*/

		test	= down_interruptible(&dev->sem);

		if (test)
		{
			/* We got a signal rather than the semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		test	= down_interruptible(&dev->write.sem);
		up(&dev->sem);

		if (test)
		{
			/* We got a signal rather than the WRITE semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		/*
		*	Verify against mmap() usage.
		*/

		if ((dev->mmap.write.usage) && (buff))
		{
			/* Must be NULL when mmap() is in use. */
			status	= -EINVAL;
			up(&dev->write.sem);
			break;
		}

		// Verify the count.

		if (dev->write.sample_bits == HPDI32_IO_SAMPLE_BITS_32)
		{
			samples	= count / 4;

			if (count % 4)
			{
				// The size must be a multiple of 4 bytes.
				status	= -EINVAL;
				up(&dev->write.sem);
				break;
			}
		}
		else if (dev->write.sample_bits == HPDI32_IO_SAMPLE_BITS_16)
		{
			samples	= count / 2;

			if (count % 2)
			{
				// The size must be a multiple of 2 bytes.
				status	= -EINVAL;
				up(&dev->write.sem);
				break;
			}
		}
		else
		{
			samples	= count;
		}

		/*
		*	Prepare for the operation.
		*/

		if ((buff == NULL) && (samples > dev->write.samples))
			samples	= dev->write.samples;

		if ((filp->f_flags & O_NONBLOCK) || (dev->write.timeout <= 0))
			jit_end	= 0;
		else
			jit_end	= jiffies + dev->write.timeout * HZ;

		/*
		*	Perform the operation.
		*/

		if (	(samples <= dev->write.pio_threshold) ||
			(dev->write.mode == HPDI32_IO_MODE_PIO))
		{
			status	= _write_work(	dev,
						buff,
						samples,
						jit_end,
						pio_write_available,
						pio_write_work,
						_INT_INDEX_TX_FIFO_FULL);
		}
		else if (dev->write.mode == HPDI32_IO_MODE_DMA)
		{
			status	= _write_work(	dev,
						buff,
						samples,
						jit_end,
						dma_write_available,
						dma_write_work,
						_INT_INDEX_TX_FIFO_AF);
		}
		else if (dev->write.mode == HPDI32_IO_MODE_DMDMA)
		{
			status	= _write_work(	dev,
						buff,
						samples,
						jit_end,
						dmdma_write_available,
						dmdma_write_work,
						_INT_INDEX_TX_FIFO_AF);
		}
		else
		{
			status	= -EINVAL;
		}

		/*
		*	Clean up.
		*/

		up(&dev->write.sem);
		break;
	}

	return(status);
}


