/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 3 $ */
/* $Date: 9/28/06 5:45p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.17.0-GSC-DN/driver/io_dma.c $
 *
 * 3     9/28/06 5:45p Dnewbold
 * Added code to support the Data Size and PIO Threshold parameters, as
 * well as the HPDI32's Single Cycle Disable feature, when present.
 *
 * 2     9/01/05 1:27p Dnewbold
 * Added an event condition variable for the 2.6 kernel.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/******************************************************************************
*
*	Function:	dma_read_available
*
*	Purpose:
*
*		Determine the number of samples that can be read for a
*		standard DMA mode read operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to read.
*
******************************************************************************/

ssize_t dma_read_available(dev_data_t* dev, ssize_t samples)
{
	u32	bsr;

	bsr	= readl(dev->vaddr.gsc_bsr_32);

	if ((bsr & HPDI32_GSC_BSR_RX_NOT_EMPTY) == 0)
		samples	= 0;
	else if (samples > dev->read.samples)
		samples	= dev->read.samples;

	return(samples);
}



/******************************************************************************
*
*	Function:	dma_read_work
*
*	Purpose:
*
*		Implement the data reading portion of the standard DMA read
*		operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples read go here.
*
*		samples	The number of samples to read.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no data is available.
*
*	Returned:
*
*		>= 0	This is the number of samples read.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t dma_read_work(	dev_data_t*	dev,
			u8*		buff,
			ssize_t		samples,
			unsigned long	jit_end)
{
	ssize_t	qty;

	qty	= dma_perform(	dev,
				&dev->read.queue,
				&dev->read.condition,
				jit_end,
				_DMA_READ,
				DMA_DPR_BOARD_TO_HOST,
				DMA_MODE_STANDARD_DMA,
				dev->read.sample_bits,
				buff,
				samples);
	return(qty);
}



/******************************************************************************
*
*	Function:	dma_write_available
*
*	Purpose:
*
*		Determine the number of samples that can be written for a
*		standard DMA mode write operation. For standard DMA operations
*		the application must perform safe writes.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to write.
*
******************************************************************************/

ssize_t dma_write_available(dev_data_t* dev, ssize_t samples)
{
	u32	bsr;

	bsr	= readl(dev->vaddr.gsc_bsr_32);

	if ((bsr & HPDI32_GSC_BSR_TX_NOT_FULL) == 0)
		samples	= 0;
	else if (samples > dev->write.samples)
		samples	= dev->write.samples;

	return(samples);
}



/******************************************************************************
*
*	Function:	dma_write_work
*
*	Purpose:
*
*		Implement the data writing portion of the standard DMA write
*		operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples come from here.
*
*		samples	The number of samples to write.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no space is available.
*
*	Returned:
*
*		>= 0	This is the number of samples written.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t dma_write_work(	dev_data_t*	dev,
			const u8*	buff,
			ssize_t		samples,
			unsigned long	jit_end)
{
	ssize_t	qty;

	qty	= dma_perform(	dev,
				&dev->write.queue,
				&dev->write.condition,
				jit_end,
				_DMA_WRITE,
				DMA_DPR_HOST_TO_BOARD,
				DMA_MODE_STANDARD_DMA,
				dev->write.sample_bits,
				(unsigned long*) buff,
				samples);
	return(qty);
}



/******************************************************************************
*
*	Function:	dmdma_read_available
*
*	Purpose:
*
*		Determine the number of samples that can be read for a
*		Demand Mode DMA mode read operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to read.
*
******************************************************************************/

ssize_t dmdma_read_available(dev_data_t* dev, ssize_t samples)
{
	if (samples > dev->read.samples)
		samples	= dev->read.samples;

	return(samples);
}



/******************************************************************************
*
*	Function:	dmdma_read_work
*
*	Purpose:
*
*		Implement the data reading portion of the Demand Mode DMA read
*		operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples read go here.
*
*		samples	The number of samples to read.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no data is available.
*
*	Returned:
*
*		>= 0	This is the number of samples read.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t dmdma_read_work(dev_data_t*	dev,
			u8*		buff,
			ssize_t		samples,
			unsigned long	jit_end)
{
	u32	ae;
	u32	bcr;
	ssize_t	dmdma;
	ssize_t	pio;
	ssize_t	qty	= 0;
	u32	rar;

	if ((dev->cache.gsc_fsr & HPDI32_GSC_FSR_SCD) == 0)
	{
		// The Single Cycle Disable feature is unsupported.
		// Take no special action.
		dmdma	= samples;
		pio	= 0;
	}
	else if (dev->read.sample_bits < dev->bus_bits)
	{
		// Apply the Single Cycle Disable bit.
		// See how many samples to read via PIO.
		bcr	= readl(dev->vaddr.gsc_bcr_32);
		rar	= readl(dev->vaddr.gsc_rar_32);
		ae	= HPDI32_GSC_RAR_DECODE_EMPTY(rar);

		if ((bcr & HPDI32_GSC_BCR_SING_CYC_DIS) == 0)
		{
			// Apply the Single Cycle Disable bit.
			bcr	|= HPDI32_GSC_BCR_SING_CYC_DIS;
			writel(bcr, dev->vaddr.gsc_bcr_32);
		}

		if (samples > ae)
		{
			dmdma	= samples - ae;
			pio	= ae;
		}
		else
		{
			dmdma	= 0;
			pio	= samples;
		}

		if ((dmdma) && (dmdma < dev->read.pio_threshold))
		{
			// Revert to all PIO due to the threshold.
			dmdma	= 0;
			pio	= samples;
		}
	}
	else
	{
		// Clear the Single Cycle Disable bit.
		// Otherwise take no special action.
		dmdma	= samples;
		pio	= 0;
		bcr	= readl(dev->vaddr.gsc_bcr_32);

		if (bcr & HPDI32_GSC_BCR_SING_CYC_DIS)
		{
			// Apply the Single Cycle Disable bit.
			bcr	&= ~HPDI32_GSC_BCR_SING_CYC_DIS;
			writel(bcr, dev->vaddr.gsc_bcr_32);
		}
	}

	if (dmdma)
	{
		qty	= dma_perform(	dev,
					&dev->read.queue,
					&dev->read.condition,
					jit_end,
					_DMDMA_READ,
					DMA_DPR_BOARD_TO_HOST,
					DMA_MODE_DEMAND_MODE_DMA,
					dev->read.sample_bits,
					buff,
					dmdma);

		if (qty <= 0)
			pio	= 0;	// We timed out.
		else if (pio)
			buff	+= qty * (dev->read.sample_bits / 8);
	}

	if (pio)
		qty	+= pio_read_work(dev, buff, pio, jit_end);

	return(qty);
}



/******************************************************************************
*
*	Function:	dmdma_write_available
*
*	Purpose:
*
*		Determine the number of samples that can be written for a
*		Demand Mode DMA mode write operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to write.
*
******************************************************************************/

ssize_t dmdma_write_available(dev_data_t* dev, ssize_t samples)
{
	if (samples > dev->write.samples)
		samples	= dev->write.samples;

	return(samples);
}



/******************************************************************************
*
*	Function:	dmdma_write_work
*
*	Purpose:
*
*		Implement the data writing portion of the Demand Mode DMA
*		write operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples come from here.
*
*		samples	The number of samples to write.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no space is available.
*
*	Returned:
*
*		>= 0	This is the number of samples written.
*		< 0	There was an error.
*
******************************************************************************/

ssize_t dmdma_write_work(	dev_data_t*	dev,
				const u8*	buff,
				ssize_t		samples,
				unsigned long	jit_end)
{
	u32	af;
	u32	bcr;
	ssize_t	dmdma;
	ssize_t	pio;
	ssize_t	qty	= 0;
	u32	tar;

	if ((dev->cache.gsc_fsr & HPDI32_GSC_FSR_SCD) == 0)
	{
		// The Single Cycle Dsiable feature is unsupported.
		// Take no special action.
		dmdma	= samples;
		pio	= 0;
	}
	else if (dev->write.sample_bits < dev->bus_bits)
	{
		// Apply the Single Cycle Disable bit.
		// See how many samples to write via PIO.
		bcr	= readl(dev->vaddr.gsc_bcr_32);
		tar	= readl(dev->vaddr.gsc_tar_32);
		af	= HPDI32_GSC_TAR_DECODE_FULL(tar);

		if ((bcr & HPDI32_GSC_BCR_SING_CYC_DIS) == 0)
		{
			// Apply the Single Cycle Disable bit.
			bcr	|= HPDI32_GSC_BCR_SING_CYC_DIS;
			writel(bcr, dev->vaddr.gsc_bcr_32);
		}

		if (samples > af)
		{
			dmdma	= samples - af;
			pio	= af;
		}
		else
		{
			dmdma	= 0;
			pio	= samples;
		}

		if ((dmdma) && (dmdma <= dev->write.pio_threshold))
		{
			// Revert to all PIO due to the threshold.
			dmdma	= 0;
			pio	= samples;
		}
	}
	else
	{
		// Clear the Single Cycle Disable bit.
		// Otherwise take no special action.
		dmdma	= samples;
		pio	= 0;
		bcr	= readl(dev->vaddr.gsc_bcr_32);

		if (bcr & HPDI32_GSC_BCR_SING_CYC_DIS)
		{
			// Clear the Single Cycle Disable bit.
			bcr	&= ~HPDI32_GSC_BCR_SING_CYC_DIS;
			writel(bcr, dev->vaddr.gsc_bcr_32);
		}
	}

	if (dmdma)
	{
		qty	= dma_perform(	dev,
					&dev->write.queue,
					&dev->write.condition,
					jit_end,
					_DMDMA_WRITE,
					DMA_DPR_HOST_TO_BOARD,
					DMA_MODE_DEMAND_MODE_DMA,
					dev->write.sample_bits,
					(unsigned long*) buff,
					dmdma);

		if (qty <= 0)
			pio	= 0;	// We timed out.
		else if (pio)
			buff	+= qty * (dev->write.sample_bits / 8);
	}

	if (pio)
		qty	+= pio_write_work(dev, buff, pio, jit_end);

	return(qty);
}


