/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 2 $ */
/* $Date: 9/28/06 5:45p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.17.0-GSC-DN/driver/io_pio.c $
 *
 * 2     9/28/06 5:45p Dnewbold
 * Added code to support the Data Size parameter.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/******************************************************************************
*
*	Function:	pio_read_available
*
*	Purpose:
*
*		Determine the number of samples that can be read for a PIO
*		mode read operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to read.
*
******************************************************************************/

ssize_t pio_read_available(dev_data_t* dev, ssize_t samples)
{
	unsigned long	almost;
	unsigned long	bsr;
	ssize_t		qty;
	unsigned long	rar;

	for (;;)	/* A convenience loop. */
	{
		bsr	= readl(dev->vaddr.gsc_bsr_32);

		if (dev->cache.gsc_rfsr)
		{
			/*
			*	Use the Size register.
			*/

			/*
			*	If the FIFO is full, then the number of
			*	samples is the FIFO size.
			*/

			if ((bsr & HPDI32_GSC_BSR_RX_NOT_FULL) == 0)
			{
				qty	= dev->cache.gsc_rfsr;
				break;
			}

			/*
			*	If the FIFO is Almost Full then there are at
			*	least Size - Almost Full samples.
			*/

			if ((bsr & HPDI32_GSC_BSR_RX_NOT_AF) == 0)
			{
				rar	= readl(dev->vaddr.gsc_rar_32);
				almost	= HPDI32_GSC_RAR_DECODE_FULL(rar);
				qty	= dev->cache.gsc_rfsr - almost;
				break;
			}

			/*
			*	If the FIFO is not Almost Empty then there are
			*	at least Almost Empty samples.
			*/

			if (bsr & HPDI32_GSC_BSR_RX_NOT_AE)
			{
				rar	= readl(dev->vaddr.gsc_rar_32);
				qty	= HPDI32_GSC_RAR_DECODE_EMPTY(rar);
				break;
			}

			/*
			*	If the FIFO is not Empty then there is at
			*	least one sample.
			*/

			if (bsr & HPDI32_GSC_BSR_RX_NOT_EMPTY)
			{
				qty	= 1;
				break;
			}

			/*
			*	The FIFO is empty.
			*/

			qty	= 0;
			break;
		}

		/*
		*	The SIZE registers are not available.
		*/

		/*
		*	If the FIFO is not Almost Empty then there are at
		*	leat Almost Empty samples.
		*/

		if (bsr & HPDI32_GSC_BSR_RX_NOT_AE)
		{
			rar	= readl(dev->vaddr.gsc_rar_32);
			qty	= HPDI32_GSC_RAR_DECODE_EMPTY(rar);
			break;
		}

		/*
		*	If the FIFO is not Empty then there are at least one
		*	sample.
		*/

		if (bsr & HPDI32_GSC_BSR_RX_NOT_EMPTY)
		{
			qty	= 1;
			break;
		}

		/*
		*	The FIFO is empty.
		*/

		qty	= 0;
		break;
	}

	if (qty > samples)
		qty	= samples;

	return(qty);
}



/******************************************************************************
*
*	Function:	pio_read_work
*
*	Purpose:
*
*		Implement the data reading portion of the PIO read operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples read go here.
*
*		samples	The number of samples to read.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no data is available.
*
*	Returned:
*
*		>= 0	This is the number of samples read.
*
******************************************************************************/

ssize_t pio_read_work(	dev_data_t*	dev,
			u8*		buff,
			ssize_t		samples,
			unsigned long	jit_end)
{
	VADDR_T	fdr	= dev->vaddr.gsc_fdr_32;
	u8*	pu8;
	u16*	pu16;
	u32*	pu32;
	ssize_t	qty	= samples;

	if (jit_end == 0)
		jit_end	= 0xFFFFFFFF;

	switch (dev->read.sample_bits)
	{
		case HPDI32_IO_SAMPLE_BITS_8:

			pu8	= (u8*) buff;

			for (; samples; )
			{
				samples--;
				pu8[0]	= readb(fdr);
				pu8++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;

		case HPDI32_IO_SAMPLE_BITS_16:

			pu16	= (u16*) buff;

			for (; samples; )
			{
				samples--;
				pu16[0]	= readw(fdr);
				pu16++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;

		case HPDI32_IO_SAMPLE_BITS_32:

			pu32	= (u32*) buff;

			for (; samples; )
			{
				samples--;
				pu32[0]	= readl(fdr);
				pu32++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;
	}

	qty	-= samples;
	return(qty);
}



/******************************************************************************
*
*	Function:	pio_write_available
*
*	Purpose:
*
*		Determine the number of samples that can be written for a PIO
*		mode write operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		samples	The number of samples desired.
*
*	Returned:
*
*		>= 0	This is the number of samples to read.
*
******************************************************************************/

ssize_t pio_write_available(dev_data_t* dev, ssize_t samples)
{
	unsigned long	bsr;
	ssize_t		qty;
	unsigned long	tar;

	bsr	= readl(dev->vaddr.gsc_bsr_32);

	if (bsr & HPDI32_GSC_BSR_TX_NOT_AF)
	{
		/* There are at least Almost Full spaces available. */
		tar	= readl(dev->vaddr.gsc_tar_32);
		qty	= HPDI32_GSC_TAR_DECODE_FULL(tar);
	}
	else if (bsr & HPDI32_GSC_BSR_TX_NOT_FULL)
	{
		/* The Tx FIFO isn't full. */
		qty	= 1;
	}
	else
	{
		/* The Tx FIFO is full. */
		qty	= 0;
	}

	if (qty > samples)
		qty	= samples;

	return(qty);
}



/******************************************************************************
*
*	Function:	pio_write_work
*
*	Purpose:
*
*		Implement the data writing portion of the PIO write operation.
*
*	Arguments:
*
*		dev	The device data structure.
*
*		buff	The samples read go here.
*
*		samples	The number of samples to read.
*
*		jit_end	End at this point in time (in jiffies). This is zero
*			if we're to end as soon at no data is available.
*
*	Returned:
*
*		>= 0	This is the number of samples written.
*
******************************************************************************/

ssize_t pio_write_work(	dev_data_t*	dev,
			const u8*	buff,
			ssize_t		samples,
			unsigned long	jit_end)
{
	VADDR_T		fdr	= dev->vaddr.gsc_fdr_32;
	const u8*	pu8;
	const u16*	pu16;
	const u32*	pu32;
	ssize_t		qty	= samples;

	if (jit_end == 0)
		jit_end	= 0xFFFFFFFF;

	switch (dev->write.sample_bits)
	{
		case HPDI32_IO_SAMPLE_BITS_8:

			pu8	= (const u8*) buff;

			for (; samples; )
			{
				samples--;
				writeb(pu8[0], fdr);
				pu8++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;

		case HPDI32_IO_SAMPLE_BITS_16:

			pu16	= (const u16*) buff;

			for (; samples; )
			{
				samples--;
				writew(pu16[0], fdr);
				pu16++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;

		case HPDI32_IO_SAMPLE_BITS_32:

			pu32	= (const u32*) buff;

			for (; samples; )
			{
				samples--;
				writel(pu32[0], fdr);
				pu32++;

				if (jiffies > jit_end)
				{
					/* We've times out. */
					break;
				}
			}

			break;
	}

	qty	-= samples;
	return(qty);
}



