/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 1 $ */
/* $Date: 1/14/05 8:04a $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32_Linux_Ver.1.09.0-GSC-DN/Linux 2.2, 2.4 kernels/driver/ioctl.c $
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/* typedefs	**************************************************************/

typedef	int (*dev_ioctl_service_t)(dev_data_t* dev, void* arg);

typedef struct
{
	unsigned long		cmd;
	dev_ioctl_service_t	func;
} dev_ioctl_t;



/* prototypes	**************************************************************/

static int	_debug_data_get(dev_data_t* dev, hpdi32_debug_data_t* arg);
static int	_driver_info_get(dev_data_t*, hpdi32_driver_info_t* arg);
static int	_no_command(dev_data_t* dev, void* arg);



/* variables	**************************************************************/

/*
*	IOCTL services must be entered into this table to be supported by
*	the driver. The services MUST appear in the table in the position
*	corresponding to its service number (which is encoded in the IOCTL
*	macro).
*/

static const dev_ioctl_t	_ioctl_list[]	=
{
	{
		/* cmd	*/	HPDI32_IOCTL_NO_COMMAND,
		/* func	*/	(dev_ioctl_service_t) _no_command
	},
	{
		/* cmd	*/	HPDI32_IOCTL_DEBUG_DATA_GET,
		/* func	*/	(dev_ioctl_service_t) _debug_data_get
	},
	{
		/* cmd	*/	HPDI32_IOCTL_DRIVER_INFO_GET,
		/* func	*/	(dev_ioctl_service_t) _driver_info_get
	},
	{
		/* cmd	*/	HPDI32_IOCTL_REG_READ,
		/* func	*/	(dev_ioctl_service_t) reg_read_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_REG_WRITE,
		/* func	*/	(dev_ioctl_service_t) reg_write_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_REG_MOD,
		/* func	*/	(dev_ioctl_service_t) reg_mod_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_IO_CONFIG,
		/* func	*/	(dev_ioctl_service_t) io_config_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_INT_NOTIFY,
		/* func	*/	(dev_ioctl_service_t) irq_notify_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_INT_STATUS,
		/* func	*/	(dev_ioctl_service_t) irq_status_ioctl
	},
	{
		/* cmd	*/	HPDI32_IOCTL_MMAP_INFO,
		/* func	*/	(dev_ioctl_service_t) mmap_info_ioctl
	}
};



/******************************************************************************
*
*	Function:	_debug_data_get
*
*	Purpose:
*
*		Implement the Get Debug Data IOCTL service. Copy over the
*		device data, then clear it out, plus copy over the driver
*		data, but without clearing it.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _debug_data_get(dev_data_t* dev, hpdi32_debug_data_t* arg)
{
	memcpy(&arg->device, &dev->debug, sizeof(arg->device));
	memset(&dev->debug, 0, sizeof(dev->debug));
	memcpy(&arg->driver, &global.debug, sizeof(arg->driver));
	return(0);
}



/******************************************************************************
*
*	Function:	_driver_info_get
*
*	Purpose:
*
*		Implement the Get Driver Information IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _driver_info_get(dev_data_t* dev, hpdi32_driver_info_t* arg)
{
	memcpy(arg, &global.driver_info, sizeof(global.driver_info));
	return(0);
}



/******************************************************************************
*
*	Function:	_no_command
*
*	Purpose:
*
*		Implement the No Command IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

static int _no_command(dev_data_t* dev, void* arg)
{
	return(0);
}



/******************************************************************************
*
*	Function:	ioctl
*
*	Purpose:
*
*		Implement the ioctl() procedure.
*
*	Arguments:
*
*		inode	This is the device inode structure.
*
*		filp	This is the file structure.
*
*		cmd	The desired IOCTL service.
*
*		arg	The optional service specific argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem.
*
******************************************************************************/

int ioctl(	struct inode*	inode,
		struct file*	filp,
		unsigned int	cmd,
		unsigned long	arg)
{
	union
	{
		hpdi32_debug_data_t	debug_data;
		hpdi32_driver_info_t	driver_info;
		hpdi32_io_config_t	io_config;
		hpdi32_interrupt_t	irq_data;
		hpdi32_mmap_t		mmap;
		hpdi32_reg_t		reg;
	} buf;

	dev_data_t*	dev;
	unsigned long	dir;
	unsigned long	index;
	unsigned long	size;
	int		status;
	int		test;
	unsigned long	type;
	unsigned long	ul;

	for (;;)	/* We'll use a loop for convenience. */
	{
		dev	= dev_locate(inode);

		if (dev == NULL)
		{
			/* The referenced device doesn't exist. */
			status	= -ENODEV;
			break;
		}

		test	= down_interruptible(&dev->sem);

		if (test)
		{
			/* We didn't get the semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		type	= _IOC_TYPE(cmd);

		if (type != 's')
		{
			/* The IOCTL code isn't ours. */
			status	= -EINVAL;
			up(&dev->sem);
			break;
		}

		index	= _IOC_NR(cmd);

		if (index >= ARRAY_ELEMENTS(_ioctl_list))
		{
			/* The IOCTL service is unrecognized. */
			status	= -EINVAL;
			up(&dev->sem);
			break;
		}

		if (_ioctl_list[index].func == NULL)
		{
			/* The IOCTL service is unimplemented. */
			status	= -EINVAL;
			up(&dev->sem);
			break;
		}

		size	= _IOC_SIZE(cmd);

		if (size != _IOC_SIZE(_ioctl_list[index].cmd))
		{
			/* The IOCTL data size is invalid. */
			status	= -EINVAL;
			up(&dev->sem);
			break;
		}

		dir	= _IOC_DIR(cmd);

		if (dir & _IOC_WRITE)
		{
			ul	= access_ok(VERIFY_READ, arg, size);

			if (ul == 0)
			{
				/* We can't read from the user's memory. */
				status	= -EFAULT;
				up(&dev->sem);
				break;
			}

			ul	= copy_from_user(&buf, (void*) arg, size);

			if (ul)
			{
				/* We couldn't read from the user's memory. */
				status	= -EFAULT;
				up(&dev->sem);
				break;
			}
		}

		if (dir & _IOC_READ)
		{
			ul	= access_ok(VERIFY_WRITE, arg, size);

			if (ul == 0)
			{
				/* We can't write to the user's memory. */
				status	= -EFAULT;
				up(&dev->sem);
				break;
			}
		}

		/* Process the IOCTL command. */
		status	= _ioctl_list[index].func(dev, &buf);

		if (status)
		{
			/* The service failed. */
			up(&dev->sem);
			break;
		}

		if (dir & _IOC_READ)
		{
			ul	= copy_to_user((void*) arg, &buf, size);

			if (ul)
			{
				/* We couldn't write to the user's memory. */
				status	= -EFAULT;
				up(&dev->sem);
				break;
			}
		}

		status	= 0;
		up(&dev->sem);
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	ioctl_start
*
*	Purpose:
*
*		Perform any work needed when starting the driver.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		0	All went well.
*		! 0	There was a problem.
*
******************************************************************************/

int ioctl_start(void)
{
	int	i;

	i	= global.dev_qty ? ARRAY_ELEMENTS(_ioctl_list) : 0;

	for (; i--;)
	{
		if (_IOC_NR(_ioctl_list[i].cmd) != i)
		{
			i = 1;
			printk(	"<1>%s: IOCTL list is out of order.\n",
				MODULE_NAME);
			break;
		}
	}

	i	= i < 0 ? 0 : i;
	return(i);
}



