/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 7 $ */
/* $Date: 8/16/06 12:00p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.16.0-GSC-DN/driver/irq.c $
 *
 * 7     8/16/06 12:00p Dnewbold
 * Additional kernel porting.
 *
 * 6     5/30/06 1:15p Dnewbold
 * Added irq_intcsr_mod() to control modifications to the PLX INTCSR.
 *
 * 5     4/17/06 5:09p Dnewbold
 * Corrected a bug in the isr.
 *
 * 4     2/27/06 10:34a Dnewbold
 * Correct a bug - the lock wasn't released when not our interrupt.
 *
 * 3     9/01/05 1:28p Dnewbold
 * Added an event condition variable for the 2.6 kernel.
 *
 * 2     1/25/05 3:46p Dnewbold
 * More modularization work, plus some housekeeping updates.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/* variables	**************************************************************/

static const struct
{
	unsigned long	mask;	/* The interrupt bit. */
	unsigned long	el;	/* Edge or Level triggered. */
	unsigned long	hl;	/* High or Low triggered. */
} irq_gsc[] =
{
	/*  interrupt	   mask		el		hl		*/
	{/* Frame Start	*/ 0x00000001,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Frame End	*/ 0x00000002,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Command D1	*/ 0x00000004,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Command D2	*/ 0x00000008,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Command D3	*/ 0x00000010,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Command D4	*/ 0x00000020,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Command D5	*/ 0x00000040,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Command D6	*/ 0x00000080,	_INT_EL_EDGE,	_INT_HL_HIGH	},

	/* We're using TRUE logic states here, not the inverted states. */
	{/* Tx Empty	*/ 0x00000100,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Tx Low	*/ 0x00000200,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Tx High	*/ 0x00000400,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Tx Full	*/ 0x00000800,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Rx Empty	*/ 0x00001000,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Rx Low	*/ 0x00002000,	_INT_EL_EDGE,	_INT_HL_HIGH	},
	{/* Rx High	*/ 0x00004000,	_INT_EL_EDGE,	_INT_HL_LOW	},
	{/* Rx Full	*/ 0x00008000,	_INT_EL_EDGE,	_INT_HL_LOW	}
};



/******************************************************************************
*
*	Function:	_irq_notify
*
*	Purpose:
*
*		Implement interrupt notification for the given interrupt(s)
*		according to the user's request. Since all interrupts are
*		PCI interrupts, we handle it here automatically.
*
*	Arguments:
*
*		dev	The device of interest.
*
*		gsc	Is this a GSC spefic interrupt?
*
*		mask	The interrupt bits per the class indicates by "gsc".
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _irq_notify(dev_data_t* dev, int gsc, u32 mask)
{
	u32	pci;

	for (;;)	/* A convenience loop. */
	{
		pci	= dev->irq.user_other & HPDI32_INT_PCI;

		if (gsc)
		{
			/* A GSC specific interrupt has occurred. */
			mask	&= dev->irq.user_gsc;

			if ((mask == 0) && (pci == 0))
			{
				/* Notification has not been requested. */
				break;
			}

			dev->irq.status_gsc	|= mask;
			dev->irq.status_other	|= pci;
		}
		else
		{
			/* A non-GSC specific interrupt has occurred. */
			mask	&= dev->irq.user_other;

			if ((mask == 0) && (pci == 0))
			{
				/* Notification has not been requested. */
				break;
			}

			dev->irq.status_other	|= mask | pci;
		}

		if (dev->irq.async)
			KILL_FASYNC(dev->irq.async, SIGIO, POLL_IN);

		break;
	}
}



/******************************************************************************
*
*	Function:	irq_access_lock
*
*	Purpose:
*
*		Apply a locking mechanism to prevent simultaneous access to
*		the device's IRQ substructure.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		isr	Is this an ISR based request?
*
*	Returned:
*
*		0	The request was satisfied.
*		else	There was a problem.
*
******************************************************************************/

int irq_access_lock(dev_data_t* dev, int isr)
{
	int	status;

	if (isr)
	{
		status	= 0;
#ifdef __SMP__
#else
#endif
	}
	else
	{
		status	= down_interruptible(&dev->irq.sem);

		if (status == 0)
			disable_irq(dev->irq.number);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	irq_access_unlock
*
*	Purpose:
*
*		Remove the locking mechanism that prevented simultaneous
*		access to the device's IRQ substructure.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		isr	Is this an ISR based request?
*
*	Returned:
*
*		None.
*
******************************************************************************/

void irq_access_unlock(dev_data_t* dev, int isr)
{
	if (isr)
	{
#ifdef __SMP__
#else
#endif
	}
	else
	{
		enable_irq(dev->irq.number);
		up(&dev->irq.sem);
	}
}



/******************************************************************************
*
*	Function:	irq_gsc_acquire
*
*	Purpose:
*
*		Try to gain access to the specified interrupt type, and with
*		the given configuration. If the interrupt is already in use or
*		the desired configuration is not possible, then the request
*		will fail. If the request is successful and a wait queue is
*		provided, then we go to sleep waiting for the event.
*
*		NOTES:
*
*		With GSC interrupts it is possible for the requested interrupt
*		to occure before the current task goes to sleep. This is
*		possible, for example, because we could be waiting for space
*		to become available in the Tx FIFO which could happen between
*		the time the caller checked and when we set the interrupt. So,
*		to make sure things work in these cases, we take manual steps
*		to insure that the current task can be awoken by the interrupt
*		before the task actually sleeps and before the interrupt is
*		enabled.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		queue	The wait queue to add this event to. This is for
*			driver requests only and cannot be NULL.
*
*		cond	The condition variable for the 2.6 kernel
*			event waiting and resuming feature.
*
*		jit_end	We are supposed to timeout at this time.  This is for
*			driver requests only.
*
*		index	The index of the given interrupt type.
*
*		el	Is Edge or Level triggiering desired? This is for
*			driver requests only.
*
*		hl	Is High or Low triggering desired? This is for
*			driver requests only.
*
*	Returned:
*
*		IRQ_GSC_FAIL		Request failed.
*		IRQ_GSC_TIME_SIG	Got a signal or timed out.
*		IRQ_GSC_IRQ		Got the interrupt.
*
******************************************************************************/

int irq_gsc_acquire(	dev_data_t* 		dev,
			WAIT_QUEUE_HEAD_T*	queue,
			int*			cond,
			unsigned long		jit_end,
			unsigned int		index,
			int			el,
			int			hl)
{
	unsigned long		icr;
	unsigned long		ielr;
	unsigned long		ihlr;
	unsigned long		mask;
	int			status;
	unsigned long		timeout;
	unsigned long		tmp;
	WAIT_QUEUE_ENTRY_T	wait;

	mask	= irq_gsc[index].mask;
	status	= irq_access_lock(dev, FALSE);

	if (status == 0)
		status	= ~IRQ_GSC_FAIL;
	else
		status	= IRQ_GSC_FAIL;

	for (; status != IRQ_GSC_FAIL;)	/* A convenience loop. */
	{
		if ((dev->irq.driver_gsc & mask) || (dev->irq.user_gsc & mask))
		{
			/* The interrupt is already in use. */
			status	= IRQ_GSC_FAIL;
			irq_access_unlock(dev, FALSE);
			break;
		}

		if (	(dev->cache.gsc_fsr & HPDI32_GSC_FSR_IELR_IHLR) ||
			(	(irq_gsc[index].el == el) &&
				(irq_gsc[index].hl == hl)))
		{
			/* Let the driver have it. */
			dev->irq.driver_gsc	|= mask;
			icr	= readl(dev->vaddr.gsc_icr_32);
			ielr	= readl(dev->vaddr.gsc_ielr_32);
			ihlr	= readl(dev->vaddr.gsc_ihlr_32);

			if (el == _INT_EL_EDGE)
				ielr	|= mask;
			else
				ielr	&= ~mask;

			if (hl == _INT_HL_HIGH)
				ihlr	|= mask;
			else
				ihlr	&= ~mask;

			/*
			*	Clear, then configure and enable the interrupt.
			*	We clear then set it to make sure it occurs
			*	even in cases where the interrupt is already
			*	active.
			*/

			icr	&= ~mask;
			writel(icr, dev->vaddr.gsc_icr_32);	/* clear */
			writel(ielr, dev->vaddr.gsc_ielr_32);
			writel(ihlr, dev->vaddr.gsc_ihlr_32);
			icr	|= mask;
			writel(icr, dev->vaddr.gsc_icr_32);	/* enable */

			/*
			*	Do the magic!
			*/

			memset(&wait, 0, sizeof(wait));
			WAIT_QUEUE_ENTRY_INIT(&wait, current);	/* NOTES */
			SET_CURRENT_STATE(TASK_INTERRUPTIBLE);	/* NOTES */
			add_wait_queue(queue, &wait);		/* NOTES */
			irq_access_unlock(dev, FALSE);	/* Let INT fly. */
			timeout	= jit_end - jiffies;
			EVENT_WAIT_IRQ_TO(queue, cond[0], timeout);
			remove_wait_queue(queue, &wait);	/* NOTES */
			SET_CURRENT_STATE(TASK_RUNNING);	/* NOTES */

			/*
			*	Clean up.
			*/

			tmp	= readl(dev->vaddr.gsc_icr_32);

			if (tmp & mask)
				status	= IRQ_GSC_TIME_SIG;
			else
				status	= IRQ_GSC_IRQ;

			icr	&= ~mask;
			writel(icr, dev->vaddr.gsc_icr_32);	/* clear */
			dev->irq.driver_gsc	&= ~mask;
			break;
		}

		/* The request can't be satisfied. */
		status	= IRQ_GSC_FAIL;
		irq_access_unlock(dev, FALSE);
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	irq_intcsr_mod
*
*	Purpose:
*
*		Modify the PLX INTCSR register with restricted access. THIS IS
*		FOR NON ISR_ACCESSES ONLY!
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		value	This is the value to apply.
*
*		mask	This is the set of bits to modify.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void irq_intcsr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	u32	intcsr;
	int	status;

	status	= irq_access_lock(dev, FALSE);
	intcsr	= readl(dev->vaddr.plx_intcsr_32);
	intcsr	= (intcsr & ~mask) | (value & mask);
	writel(intcsr, dev->vaddr.plx_intcsr_32);

	if (status == 0)
		irq_access_unlock(dev, FALSE);
}



/******************************************************************************
*
*	Function:	irq_isr_common
*
*	Purpose:
*
*		Service an interrupt. For safety sake, each detected interrupt
*		is cleared then disabled.
*
*	Arguments:
*
*		irq	The interrupt number.
*
*		dev_id	The private data we've associated with the IRQ.
*
*		regs	Unused.
*
*	Returned:
*
*		FALSE	The interrupt was NOT ours.
*		TRUE	The interrupt was ours.
*
******************************************************************************/

int irq_isr_common(int irq, void* dev_id)
{
	dev_data_t*	dev	= dev_id;
	dev_dma_t*	dma;
	int		handled	= TRUE;
	unsigned long	icr;	/* GSC Interrupt Control */
	u32		intcsr;	/* PLX Interrupt Control/Status */
	unsigned long	isr;	/* GSC */
	unsigned long	mask;

	irq_access_lock(dev, TRUE);

	for (;;)
	{
		intcsr	= readl(dev->vaddr.plx_intcsr_32);

		if (	(intcsr & INTCSR_DMA_0_IRQ_ENABLE) &&
			(intcsr & INTCSR_DMA_0_IRQ_ACTIVE))
		{
			/* A thread is blocked on DMA channel 0. */
			dma	= &dev->dma.channel[0];

			/* Clear the DONE interrupt. */
			writeb(DMA_CSR_CLEAR, dev->vaddr.plx_dmacsr0_8);

			/* Route the DMA 0 interrupt to local space. */
			writel(0, dma->vaddr.mode_32);

			/* Disable the DMA 0 interrupt. */
			intcsr	&= ~dma->irq_enable;
			writel(intcsr, dev->vaddr.plx_intcsr_32);

			/* Resume the blocked task. */
			EVENT_RESUME_IRQ(dma->queue, dma->condition[0]);

			/* Notify the application. */
			_irq_notify(dev, FALSE, HPDI32_INT_DMA_0_DONE);
			break;
		}

		if (	(intcsr & INTCSR_DMA_1_IRQ_ENABLE) &&
			(intcsr & INTCSR_DMA_1_IRQ_ACTIVE))
		{
			/* A thread is blocked on DMA channel 1. */
			dma	= &dev->dma.channel[1];

			/* Clear the DONE interrupt. */
			writeb(DMA_CSR_CLEAR, dev->vaddr.plx_dmacsr1_8);

			/* Route the DMA 1 interrupt to local space. */
			writel(0, dma->vaddr.mode_32);

			/* Disable the DMA 1 interrupt. */
			intcsr	&= ~dma->irq_enable;
			writel(intcsr, dev->vaddr.plx_intcsr_32);

			/* Resume the blocked task. */
			EVENT_RESUME_IRQ(dma->queue, dma->condition[0]);

			/* Notify the application. */
			_irq_notify(dev, FALSE, HPDI32_INT_DMA_1_DONE);
			break;
		}

		isr	= readl(dev->vaddr.gsc_isr_32);
		icr	= readl(dev->vaddr.gsc_icr_32);
		mask	= icr & isr;	/* Those that we've caught. */

		if (mask & HPDI32_INT_GSC_RX_FIFO_MASK & dev->irq.driver_gsc)
		{
			/* A READ thread is blocked on a GSC interrupt. */
			/* Clear the interrupt. */
			mask	&= HPDI32_INT_GSC_RX_FIFO_MASK
				& dev->irq.driver_gsc;
			writel(mask, dev->vaddr.gsc_isr_32);

			/* Disable the interrupt. */
			icr	&= ~mask;
			writel(icr, dev->vaddr.gsc_icr_32);

			/* Resume the blocked task. */
			EVENT_RESUME_IRQ(	&dev->read.queue,
						dev->read.condition);

			/* Notify the application. */
			_irq_notify(dev, TRUE, mask);
			break;
		}

		if (mask & HPDI32_INT_GSC_TX_FIFO_MASK & dev->irq.driver_gsc)
		{
			/* A WRITE thread is blocked on a GSC interrupt. */
			/* Clear the interrupt. */
			mask	&= HPDI32_INT_GSC_TX_FIFO_MASK
				& dev->irq.driver_gsc;
			writel(mask, dev->vaddr.gsc_isr_32);

			/* Disable the interrupt. */
			icr	&= ~mask;
			writel(icr, dev->vaddr.gsc_icr_32);

			/* Resume the blocked task. */
			EVENT_RESUME_IRQ(	&dev->write.queue,
						dev->write.condition);

			/* Notify the application. */
			_irq_notify(dev, TRUE, mask);
			break;
		}

		if (mask)
		{
			/* An application configured interrupt has occurred. */
			/* Clear the interrupt. */
			writel(mask, dev->vaddr.gsc_isr_32);

			/* Disable the interrupt. */
			writel(icr & ~mask, dev->vaddr.gsc_icr_32);

			/* Notify the application. */
			_irq_notify(dev, TRUE, mask);
			break;
		}

		/*
		*	Well, it looks like an interrupt for another board
		*	sharing our interrupt line.
		*/

		handled	= FALSE;
		break;
	}

	irq_access_unlock(dev, TRUE);
	return(handled);
}



/******************************************************************************
*
*	Function:	irq_notify_ioctl
*
*	Purpose:
*
*		Implement the Interrupt Notify IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int irq_notify_ioctl(dev_data_t* dev, hpdi32_interrupt_t* arg)
{
	u32	intcsr;
	int	status;

	status	= irq_access_lock(dev, FALSE);

	if (status)
	{
		/* We couldn't get access. */
		status	= -ERESTARTSYS;
	}
	else
	{
		/* Notify the application of these. */
		dev->irq.user_gsc	= arg->notify.gsc;
		dev->irq.user_other	= arg->notify.other;

		/* The application can control these. */
		arg->status.gsc		= dev->irq.user_gsc
					& ~dev->irq.driver_gsc;
		arg->status.other	= 0;

		/*
		*	Enable the PCI interrupt in case we disabled it
		*	because of an unexpected non-GSC interrupt.
		*/

		intcsr	= readl(dev->vaddr.plx_intcsr_32);
		intcsr	|= INTCSR_PCI_INT_ENABLE;
		writel(intcsr, dev->vaddr.plx_intcsr_32);

		/* We're done. */
		status	= 0;
		irq_access_unlock(dev, FALSE);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	irq_request
*
*	Purpose:
*
*		Request that this deivce be allotted its assigned IRQ.
*
*	Arguments:
*
*		dev	The device of interest.
*
*	Returned:
*
*		0	All went well.
*		1	There was an error.
*
******************************************************************************/

int irq_request(dev_data_t* dev)
{
	int	status;

	status	= request_irq(	dev->pci->irq,
				irq_isr,
				SA_SHIRQ,
				MODULE_NAME,
				dev);

	if (status)
		status	= 1;
	else
		dev->irq.number	= dev->pci->irq;

	return(status);
}



/******************************************************************************
*
*	Function:	irq_status_ioctl
*
*	Purpose:
*
*		Implement the Interrupt Status IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int irq_status_ioctl(dev_data_t* dev, hpdi32_interrupt_t* arg)
{
	int	status;

	status	= irq_access_lock(dev, FALSE);

	if (status)
	{
		/* We couldn't get access. */
		status	= -ERESTARTSYS;
	}
	else
	{
		/* Report the accumulated status. */
		arg->notify.gsc		= dev->irq.user_gsc;
		arg->notify.other	= dev->irq.user_other;
		arg->status.gsc		= dev->irq.status_gsc;
		arg->status.other	= dev->irq.status_other;

		/* Clear the accumulated status. */
		dev->irq.status_gsc	= 0;
		dev->irq.status_other	= 0;

		/* We're done. */
		status	= 0;
		irq_access_unlock(dev, FALSE);
	}

	return(status);
}



