/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 5 $ */
/* $Date: 5/30/06 1:29p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32/HPDI32_Linux_Ver.1.15.0-GSC-DN/driver/mmap.c $
 *
 * 5     5/30/06 1:29p Dnewbold
 * Removed the VM_RESERVED flag to prevent perminant memory loss.
 *
 * 4     2/27/06 10:39a Dnewbold
 * Updatred for new I/O memory allocation mechanism.
 *
 * 3     2/01/06 3:01p Dnewbold
 * Renamed  KMALLOC_LIMIT to  MEM_ALLOC_LIMIT.
 *
 * 2     1/25/05 3:46p Dnewbold
 * More modularization work, plus some housekeeping updates.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/******************************************************************************
*
*	Function:	_mmap_close_work
*
*	Purpose:
*
*		Perform the common work needed for an mmap() close operation.
*
*	Arguments:
*
*		dev	The device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _mmap_close_work(dev_data_t* dev)
{
	dev->mmap.usage--;
	MOD_DEC_USE_COUNT;

	if (dev->mmap.usage == 0)
	{
		dev->mmap.read.usage	= 0;
		dev->mmap.write.usage	= 0;
	}
}



/******************************************************************************
*
*	Function:	_mmap_close
*
*	Purpose:
*
*		Implement the mmap close() procedure.
*
*	Arguments:
*
*		vma	The virtual memory area of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _mmap_close(struct vm_area_struct* vma)
{
	dev_data_t*	dev;
	int		test;

	dev	= dev_locate(vma->vm_file->f_dentry->d_inode);

	if (dev)
	{
		/*
		*	Due to our limited activity, we'll do the work even
		*	if we get a signal rather than the semaphore.
		*/

		test	= down_interruptible(&dev->sem);
		_mmap_close_work(dev);

		if (test == 0)
			up(&dev->sem);
	}
}



/******************************************************************************
*
*	Function:	_mmap_open
*
*	Purpose:
*
*		Implement the mmap open() procedure.
*
*	Arguments:
*
*		vma	The virtual memory area of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _mmap_open(struct vm_area_struct* vma)
{
	dev_data_t*	dev;
	int		test;

	dev	= dev_locate(vma->vm_file->f_dentry->d_inode);

	if (dev)
	{
		/*
		*	Increment our usage count, and make sure the data
		*	transfer buffers are marked as reserved so the MM will
		*	leave them along.
		*/

		test	= down_interruptible(&dev->sem);
		dev->mmap.usage++;
		MOD_INC_USE_COUNT;

		if (test == 0)
			up(&dev->sem);
	}
}



/******************************************************************************
*
*	Function:	mmap
*
*	Purpose:
*
*		Implement the mmap() procedure.
*
*	Arguments:
*
*		filp	The kernel file structure for the open device.
*
*		vma	The virtual memory area of interest.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was an error.
*
******************************************************************************/

int mmap(struct file* filp, struct vm_area_struct* vma)
{
	dev_data_t*	dev;
	unsigned long	offset;
	unsigned long	phys_adrs;
	int		status;
	unsigned long	size;
	int		test;

	for (;;)	/* We'll use a loop for convenience. */
	{
		/*
		*	Validate some of the parameters.
		*/

		dev	= (dev_data_t*) filp->private_data;

		if (dev == NULL)
		{
			/* The referenced device doesn't exist. */
			status	= -ENODEV;
			break;
		}

		/*
		*	Gain the neccessary access.
		*/

		test	= down_interruptible(&dev->sem);

		if (test)
		{
			/* We got a signal rather than the semaphore. */
			status	= -ERESTARTSYS;
			break;
		}

		/*
		*	See which memory area this request applies to.
		*/

		offset	= VM_OFFSET_BYTES(vma);
		size	= vma->vm_end - vma->vm_start;

		if (	(offset == dev->mmap.regs.offset) &&
			(size == dev->mmap.regs.pg_size))
		{
			/* This is the GSC register block. */
			vma->vm_flags	|= VM_IO | VM_RESERVED;
			status		= REMAP_PAGE_RANGE(
						vma,
						vma->vm_start,
						dev->gsc_mem.phys_adrs,
						GSC_REG_BLOCK_SIZE,
						vma->vm_page_prot);

			if (status)
			{
				/* The request failed. */
				status	= -EAGAIN;
				up(&dev->sem);
			}
			else
			{
				status		= 0;
				vma->vm_file	= filp;
				vma->vm_ops	= &global.vmops;
				up(&dev->sem);
				_mmap_open(vma);
			}

			break;
		}

		if (	(offset == dev->mmap.read.offset) &&
			(size == dev->mmap.read.pg_size))
		{
			/* This is the read() buffer. */
			// The VM_RESERVED flag isn't set as this memory already has that
			// flag applied. It was applied when it was allocated. If we apply
			// it here as well, then the memory will never be freed.
			phys_adrs 	= __pa(dev->read.buf);
			status		= REMAP_PAGE_RANGE(
						vma,
						vma->vm_start,
						phys_adrs,
						size,
						vma->vm_page_prot);

			if (status)
			{
				/* The request failed. */
				status	= -EAGAIN;
				up(&dev->sem);
			}
			else
			{
				status		= 0;
				vma->vm_file	= filp;
				vma->vm_ops	= &global.vmops;
				up(&dev->sem);
				_mmap_open(vma);
			}

			break;
		}

		if (	(offset == dev->mmap.write.offset) &&
			(size == dev->mmap.write.pg_size))
		{
			/* This is the write() buffer. */
			phys_adrs 	= __pa(dev->write.buf);
			status		= REMAP_PAGE_RANGE(
						vma,
						vma->vm_start,
						phys_adrs,
						size,
						vma->vm_page_prot);

			if (status)
			{
				/* The request failed. */
				status	= -EAGAIN;
				up(&dev->sem);
			}
			else
			{
				status		= 0;
				vma->vm_file	= filp;
				vma->vm_ops	= &global.vmops;
				up(&dev->sem);
				_mmap_open(vma);
			}

			break;
		}

		/* The region is invalid. */
		status	= -EINVAL;
		up(&dev->sem);
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	mmap_close_dev
*
*	Purpose:
*
*		Perform any mmap work needed when closing a device.
*
*	Arguments:
*
*		dev	The device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void mmap_close_dev(dev_data_t* dev)
{
	for (; dev->mmap.usage > 0;)
		_mmap_close_work(dev);
}



/******************************************************************************
*
*	Function:	mmap_info_compute
*
*	Purpose:
*
*		Compute the set of acceptable mmap() memory areas.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void mmap_info_compute(dev_data_t* dev)
{
	u32	offset;

	/* These are the parameters used to access the GSC registers. */
	dev->mmap.regs.offset	= 0;

	if (dev->gsc_mem.phys_adrs % PAGE_SIZE)	/* Must be page alighed. */
		dev->mmap.regs.size	= 0;
	else
		dev->mmap.regs.size	= GSC_REG_BLOCK_SIZE;

	dev->mmap.regs.pg_size	= BYTES_TO_PAGE_BYTES(dev->mmap.regs.size);
	offset			= dev->mmap.regs.pg_size;

	/* These are the parameters used to access the read() buffer. */
	dev->mmap.read.offset	= offset;
	dev->mmap.read.size	= dev->read.samples * 4;

	if (dev->mmap.read.size > MEM_ALLOC_LIMIT)
		dev->mmap.read.size	= MEM_ALLOC_LIMIT;

	dev->mmap.read.pg_size	= BYTES_TO_PAGE_BYTES(dev->mmap.read.size);
	offset	+= dev->mmap.read.pg_size;

	/* These are the parameters used to access the write() buffer. */
	dev->mmap.write.offset	= offset;
	dev->mmap.write.size	= dev->write.samples * 4;

	if (dev->mmap.write.size > MEM_ALLOC_LIMIT)
		dev->mmap.write.size	= MEM_ALLOC_LIMIT;

	dev->mmap.write.pg_size	= BYTES_TO_PAGE_BYTES(dev->mmap.write.size);
}



/******************************************************************************
*
*	Function:	mmap_info_ioctl
*
*	Purpose:
*
*		Implement the mmap() information IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int mmap_info_ioctl(dev_data_t* dev, hpdi32_mmap_t* arg)
{
	arg->regs.offset	= dev->mmap.regs.offset;
	arg->regs.size		= dev->mmap.regs.size;
	arg->read.offset	= dev->mmap.read.offset;
	arg->read.size		= dev->mmap.read.size;
	arg->write.offset	= dev->mmap.write.offset;
	arg->write.size		= dev->mmap.write.size;
	return(0);
}



/******************************************************************************
*
*	Function:	mmap_nopage_common
*
*	Purpose:
*
*		Implement the common portions of the mmap nopage() procedure.
*
*	Arguments:
*
*		vma		The virtual memory area of interest.
*
*		address		The virtual address of the page being accessed.
*
*		page		We record the page pointer here.
*				This will be NULL when the mapping fails.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void mmap_nopage_common(struct vm_area_struct*	vma,
			unsigned long		address,
			struct page**		page)
{
	dev_data_t*	dev;
	unsigned long	indent;	// The distance into the allocated buffer.
	unsigned long	offset;	// The base offset per IOCTL calculation
	unsigned long	memptr;

	offset	= VM_OFFSET_BYTES(vma);
	indent	= address - vma->vm_start;
	dev	= dev_locate(vma->vm_file->f_dentry->d_inode);

	if (dev == NULL)
	{
		page[0]	= NULL;
	}
	else if (	(offset == dev->mmap.read.offset) &&
			(indent < dev->mmap.read.size))
	{
		/* This is a read() buffer access. */
		memptr	= ((unsigned long) dev->read.buf) + indent;
		page[0]	= VIRT_TO_PAGE(memptr);
		GET_PAGE(page[0]);
	}
	else if (	(offset == dev->mmap.write.offset) &&
			(indent < dev->mmap.write.size))
	{
		/* This is a write() buffer access. */
		memptr	= ((unsigned long) dev->write.buf) + indent;
		page[0]	= VIRT_TO_PAGE(memptr);
		GET_PAGE(page[0]);
	}
	else
	{
		/* This is an invalid access. */
		page[0]	= NULL;
	}
}



/******************************************************************************
*
*	Function:	mmap_setup
*
*	Purpose:
*
*		Perform any work needed to setup for the use of the mmap
*		feature.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void mmap_setup(void)
{
	global.vmops.open	= _mmap_open;
	global.vmops.close	= _mmap_close;
	global.vmops.nopage	= mmap_nopage;
	VM_OPS_UNMAP(global.vmops, mmap_unmap);
}



