/* This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL). */
/* $Revision: 2 $ */
/* $Date: 12/19/05 1:58p $ */
/* $Log: /NEW SOURCESAFE DATABASE/DIGITAL/Drivers/LINUX/HPDI32_Linux_Ver.1.12.1-GSC-DN/Linux 2.2, 2.4, 2.6 kernels/driver/reg.c $
 *
 * 2     12/19/05 1:58p Dnewbold
 * Corrected a bug in the reg_mod code.
 *
 * 1     1/14/05 8:04a Dnewbold
*/

#include "main.h"



/* variables	**************************************************************/

static const dev_region_t	_pci_region	=
{
	/* address	*/	0,	/* Required for PCI registers. */
	/* vaddr	*/	0,	/* Ignored for PCI registers. */
	/* flags	*/	0,	/* Ignored for PCI registers. */
	/* size		*/	256,	/* Required for PCI registers. */
	/* requested	*/	0	/* Ignored for PCI registers. */
};



/******************************************************************************
*
*	Function:	_reg_decode
*
*	Purpose:
*
*		Decode a register id.
*
*	Arguments:
*
*		region	This is the resource region where the register resides.
*
*		reg	This is the register of interest.
*
*		dreg	The decoded data is put here.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _reg_decode(const dev_region_t*	region,
			unsigned long		reg,
			dev_reg_t*		dreg)
{
	dreg->reg	= reg;
	dreg->region	= region;
	dreg->type	= HPDI32_REG_TYPE(reg);
	dreg->offset	= HPDI32_REG_OFFSET(reg);
	dreg->size	= HPDI32_REG_SIZE(reg);
	dreg->vaddr	= (VADDR_T) dreg->offset;
	dreg->vaddr	+= (unsigned long) region->vaddr;
}



/******************************************************************************
*
*	Function:	_reg_mem_mod
*
*	Purpose:
*
*		Perform a read-modify-write operation on a memory mapped
*		register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		dreg	The register definition of interest.
*
*		value	The value bits to apply.
*
*		mask	The mask of bits to modify.
*
*	Returned:
*
*		0	All went well.
*		1	There was a problem.
*
******************************************************************************/

static int _reg_mem_mod(dev_data_t*	dev,
			dev_reg_t*	dreg,
			unsigned long	value,
			unsigned long	mask)
{
	unsigned long	ul;

	switch (dreg->size)
	{
		default:
		case 1:	mask	&= 0xFF;
			value	&= 0xFF & mask;
			ul	= readb(dreg->vaddr);
			ul	&= ~mask;
			ul	|= value;
			writeb(ul, dreg->vaddr);
			break;

		case 2:	mask	&= 0xFFFF;
			value	&= 0xFFFF & mask;
			ul	= readw(dreg->vaddr);
			ul	&= ~mask;
			ul	|= value;
			writew(ul, dreg->vaddr);
			break;

		case 4:	mask	&= 0xFFFFFFFF;
			value	&= 0xFFFFFFFF & mask;
			ul	= readl(dreg->vaddr);
			ul	&= ~mask;
			ul	|= value;
			writel(ul, dreg->vaddr);
			break;
	}

	return(0);
}



/******************************************************************************
*
*	Function:	_reg_mem_read
*
*	Purpose:
*
*		Read a value from a memory mapped register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		dreg	The data for the register of interest.
*
*		value	The value read is recorded here.
*
*	Returned:
*
*		0	All went well.
*		1	There was a problem.
*
******************************************************************************/

static int _reg_mem_read(dev_data_t* dev, dev_reg_t* dreg, u32* value)
{
	switch (dreg->size)
	{
		default:
		case 1:	value[0]	= readb(dreg->vaddr);
			break;

		case 2:	value[0]	= readw(dreg->vaddr);
			break;

		case 4:	value[0]	= readl(dreg->vaddr);
			break;
	}

	return(0);
}



/******************************************************************************
*
*	Function:	_reg_mem_write
*
*	Purpose:
*
*		Write a value to a memory mapped register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		dreg	The register definition of interest.
*
*		value	The value to write. Bits outside the register's size
*			are ignored.
*
*	Returned:
*
*		0	All went well.
*		1	There was a problem.
*
******************************************************************************/

static int _reg_mem_write(	dev_data_t*	dev,
				dev_reg_t*	dreg,
				unsigned long	value)
{
	switch (dreg->size)
	{
		default:
		case 1:	writeb(value, dreg->vaddr);
			break;

		case 2:	writew(value, dreg->vaddr);
			break;

		case 4:	writel(value, dreg->vaddr);
			break;
	}

	return(0);
}



/******************************************************************************
*
*	Function:	_reg_pci_read
*
*	Purpose:
*
*		Read a value from a PCI register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		dreg	The data for the register of interest.
*
*		value	The value read is recorded here.
*
*	Returned:
*
*		0	All went well.
*		1	There was a problem.
*
******************************************************************************/

static int _reg_pci_read(dev_data_t* dev, dev_reg_t* dreg, u32* value)
{
	u8	byte;
	u32	dword;
	u16	word;

	switch (dreg->size)
	{
		default:
		case 1:	pci_read_config_byte(	dev->pci,
						(unsigned long) dreg->vaddr,
						&byte);
			value[0]	= byte;
			break;

		case 2:	pci_read_config_word(	dev->pci,
						(unsigned long) dreg->vaddr,
						&word);
			value[0]	= word;
			break;

		case 4:	pci_read_config_dword(	dev->pci,
						(unsigned long) dreg->vaddr,
						&dword);
			value[0]	= dword;
			break;
	}

	return(0);
}



/******************************************************************************
*
*	Function:	_reg_validate
*
*	Purpose:
*
*		Verify that a regiter id is valid.
*
*	Arguments:
*
*		dreg	The decoded data of interest.
*
*	Returned:
*
*		0	All went well.
*		1	There was a problem.
*
******************************************************************************/

static int _reg_validate(dev_reg_t* dreg)
{
	int		status	= 1;
	unsigned long	ul;

	for (;;)	/* We'll use a loop for convenience. */
	{
		/* Are there extranious bits in the register id? */
		ul	= HPDI32_REG_ENCODE(	dreg->type,
						dreg->size,
						dreg->offset);

		if (ul != dreg->reg)
			break;

		/* Does the register extend past the end of the region? */
		ul	= dreg->offset + dreg->size - 1;

		if (ul >= dreg->region->size)
			break;

		/* Is the register's size valid? */

		if (strchr("\001\002\004", (int) dreg->size) == NULL)
			break;

		/* Is the register properly aligned? */

		if ((dreg->size == 2) && (dreg->offset & 0x1))
			break;
		else if ((dreg->size == 4) && (dreg->offset & 0x3))
			break;

		/*
		*	We don't test the "type" since that is validated
		*	before the register is decoded.
		*/

		status	= 0;
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_mod_ioctl
*
*	Purpose:
*
*		Implement the Register Modify (read-modify-write) IOCTL
*		service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int reg_mod_ioctl(dev_data_t* dev, hpdi32_reg_t* arg)
{
	dev_reg_t	dreg;
	int		status;
	unsigned long	type	= HPDI32_REG_TYPE(arg->reg);

	switch (type)
	{
		default:
		case HPDI32_PCI_REGISTER:	/* READ ONLY! */
		case HPDI32_PLX_REGISTER:	/* READ ONLY! */

			status	= -EINVAL;
			break;

		case HPDI32_GSC_REGISTER:

			_reg_decode(&dev->gsc_mem, arg->reg, &dreg);
			status	= _reg_validate(&dreg);

			if (status == 0)
			{
				status	= _reg_mem_mod(	dev,
							&dreg,
							arg->value,
							arg->mask);
			}

			status	= status ? -EINVAL : 0;
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_read_ioctl
*
*	Purpose:
*
*		Implement the Register Read IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int reg_read_ioctl(dev_data_t* dev, hpdi32_reg_t* arg)
{
	dev_reg_t	dreg;
	int		status;
	unsigned long	type	= HPDI32_REG_TYPE(arg->reg);

	switch (type)
	{
		default:

			status	= -EINVAL;
			break;

		case HPDI32_GSC_REGISTER:

			_reg_decode(&dev->gsc_mem, arg->reg, &dreg);
			status	= _reg_validate(&dreg);

			if (status == 0)
			{
				status	= _reg_mem_read(dev,
							&dreg,
							&arg->value);
			}

			status	= status ? -EINVAL : 0;
			break;

		case HPDI32_PCI_REGISTER:

			_reg_decode(&_pci_region, arg->reg, &dreg);
			status	= _reg_validate(&dreg);

			if (status == 0)
			{
				status	= _reg_pci_read(dev,
							&dreg,
							&arg->value);
			}

			status	= status ? -EINVAL : 0;
			break;

		case HPDI32_PLX_REGISTER:

			_reg_decode(&dev->plx_mem, arg->reg, &dreg);
			status	= _reg_validate(&dreg);

			if (status == 0)
			{
				status	= _reg_mem_read(dev,
							&dreg,
							&arg->value);
			}

			status	= status ? -EINVAL : 0;
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_vaddr
*
*	Purpose:
*
*		Get the virtual address of the specified register.
*
*	Arguments:
*
*		dev	The respective device of interest.
*
*		reg	The register of interest.
*
*	Returned:
*
*		VADDR_T	The register's virtual address.
*
******************************************************************************/

VADDR_T reg_vaddr(dev_data_t* dev, unsigned long reg)
{
	dev_reg_t	dreg;
	unsigned long	type	= HPDI32_REG_TYPE(reg);

	switch (type)
	{
		default:
		case HPDI32_PCI_REGISTER:

			_reg_decode(&_pci_region, reg, &dreg);
			break;

		case HPDI32_PLX_REGISTER:

			_reg_decode(&dev->plx_mem, reg, &dreg);
			break;

		case HPDI32_GSC_REGISTER:

			_reg_decode(&dev->gsc_mem, reg, &dreg);
			break;
	}

	return(dreg.vaddr);
}



/******************************************************************************
*
*	Function:	reg_write_ioctl
*
*	Purpose:
*
*		Implement the Register Write IOCTL service.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		arg	The IOCTL service's required argument.
*
*	Returned:
*
*		0	All went well.
*		< 0	There was a problem that the value reflects.
*
******************************************************************************/

int reg_write_ioctl(dev_data_t* dev, hpdi32_reg_t* arg)
{
	dev_reg_t	dreg;
	int		status;
	unsigned long	type	= HPDI32_REG_TYPE(arg->reg);

	switch (type)
	{
		default:
		case HPDI32_PCI_REGISTER:	/* READ ONLY! */
		case HPDI32_PLX_REGISTER:	/* READ ONLY! */

			status	= -EINVAL;
			break;

		case HPDI32_GSC_REGISTER:

			_reg_decode(&dev->gsc_mem, arg->reg, &dreg);
			status	= _reg_validate(&dreg);

			if (status == 0)
			{
				status	= _reg_mem_write(	dev,
								&dreg,
								arg->value);
			}

			status	= status ? -EINVAL : 0;
			break;
	}

	return(status);
}



