/* $Revision: 4 $ */
/* $Date: 7/25/07 9:15a $ */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/utsname.h>

#include "HPDI32DocSrcLib.h"
#include "util_label.h"
#include "util_reg.h"



/*****************************************************************************
*
*	Function:	_print_str
*
*	Purpose:
*
*		Print a string with compacted white space.
*
*	Arguments:
*
*		str		The string to process and print.
*
*		eol		Any end of line sequence to print.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static void _print_str(const char* str, const char* eol)
{
	for (;;)
	{
		// Skip any leading white space.

		for (; isspace(str[0]);)
			str++;

		// Compact any white space that follows.

		if (str[0] == 0)
			break;

		for (;str[0];)
		{
			if (isspace(str[0]))
			{
				printf(" ");
				break;
			}

			printf("%c", str[0]);
			str++;
		}
	}

	if (eol)
		printf("%s", eol);
}



/*****************************************************************************
*
*	Function:	_id_board_frr_new
*
*	Purpose:
*
*		Identify the board from the Firmware Revision Register,
*		given its newer format.
*
*	Arguments:
*
*		frr		The Firmware Revision Register value.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_frr_new(__u32 frr)
{
	int		errs	= 0;
	__u32	fld;

	label("Firmware Revision");
	fld	= frr & 0xFF;
	printf("0x------%02lX\n", (long) fld);

	label("Board Version");
	fld	= (frr >> 8) & 0xFF;
	printf("0x----%02lX--\n", (long) fld);

	label("Hardware Revision");
	fld	= (frr >> 16) & 0xFF;
	printf("0x--%02lX---- ", (long) fld);

	if (fld == 0)
		printf("(no revision)\n");
	else
		printf("(Rev. %c)\n", (char) fld + 'A' - 1);

	label("Form Factor");
	fld	= (frr >> 24) & 0xF;
	printf("0x-%01lX------ ", (long) fld);

	if (fld == HPDI32_GSC_FRR_FF_PCI)
		printf("(PCI)\n");
	else if (fld == HPDI32_GSC_FRR_FF_PMC)
		printf("(PMC)\n");
	else if (fld == HPDI32_GSC_FRR_FF_CPCI)
		printf("(cPCI)\n");
	else if (fld == HPDI32_GSC_FRR_FF_PC104P)
		printf("(PC/104+)\n");
	else
		printf("(UNKNOWN)\n");

	label("Bus Size");
	fld	= (frr >> 28) & 0x1;
	printf("0x%01lX------- ", (long) fld);

	if (fld)
		printf("(64-bit)\n");
	else
		printf("(32-bit)\n");

	label("Reserved");
	fld	= (frr >> 28) & 0x2;
	printf("0x%01lX------- %s\n", (long) fld, fld ? "(UNRECOGNIZED)" : "");

	label("Updated Format");
	fld	= (frr >> 28) & 0x4;
	printf("0x%01lX------- ", (long) fld);

	if (fld)
	{
		printf("(Yes)\n");
	}
	else
	{
		errs++;
		printf("(INTERNAL ERROR) FAIL <---\n");
	}

	label("Feature Set Reg.");
	fld	= (frr >> 28) & 0x8;
	printf("0x%01lX------- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

	return(errs);
}



/*****************************************************************************
*
*	Function:	_id_board_frr_original
*
*	Purpose:
*
*		Identify the board from the Firmware Revision Register,
*		given its original format.
*
*	Arguments:
*
*		did		The PCI Device Id Register value.
*
*		frr		The Firmware Revision Register value.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_frr_original(__u32 did, __u32 frr)
{
	int		errs	= 0;
	__u32	fld;

	label("Firmware Revision");
	fld	= (frr >> 0) & 0xFF;
	printf("0x------%02lX\n", (long) fld);

	label("PCB Revision");
	fld	= (frr >> 8) & 0xFF;
	printf("0x----%02lX-- ", (long) fld);

	if (fld == 0)
		printf("(no revision)\n");
	else
		printf("(Rev. %c)\n", (char) fld + 'A' - 1);

	label("Sub ID");
	fld	= (frr >> 16) & 0xFF;
	printf("0x--%02lX---- ", (long) fld);

	if (fld == 0x0E)
		printf("(DiPhase 2)");

	printf("\n");

	if (did == HPDI32_DEVICE_ID_32)
	{
		label("Form Factor");
		fld	= (frr >> 24) & 0x1;
		printf("0x-%01lX------ ", (long) fld);

		if (fld)
			printf("(PMC)\n");
		else
			printf("(other than PMC)\n");
	}
	else
	{
		label("Firmware Bus Size");
		fld	= (frr >> 24) & 0x1;
		printf("0x-%01lX------ ", (long) fld);

		if (fld)
			printf("(64-bit)\n");
		else
			printf("(32-bit)\n");
	}

	label("Reserved");
	fld	= (frr >> 24) & 0x7E;
	printf("0x%02lX------ ", (long) fld);

	if (fld)
		printf("(UNRECOGNIZED)\n");
	else
		printf("\n");

	label("Feature Set Reg.");
	fld	= (frr >> 28) & 0x8;
	printf("0x%01lX------- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

	return(errs);
}



/*****************************************************************************
*
*	Function:	_id_board_frr
*
*	Purpose:
*
*		Identify the board from the Firmware Revision Register.
*
*	Arguments:
*
*		did		The PCI Device Id Register value.
*
*		frr		The value read from the Firmware Revision Register.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_frr(__u32 did, __u32 frr)
{
	int	errs;

	label("Firmware Rev. Reg.");
	printf("0x%08lX ", (long) frr);
	label_level_inc();

	if (frr & HPDI32_GSC_FRR_UPDATED)
	{
		printf("(updated format)\n");
		errs	= _id_board_frr_new(frr);
	}
	else
	{
		printf("(original format)\n");
		errs	= _id_board_frr_original(did, frr);
	}

	label_level_dec();
	return(errs);
}



/*****************************************************************************
*
*	Function:	_id_board_fs
*
*	Purpose:
*
*		Identify the board FIFO sizes.
*
*	Arguments:
*
*		frr		The value read from the Firmware Revision Register.
*
*		fsr		The value read from the Feature Set Register.
*
*		tfsr	The value read from the Tx FIFO Size Register
*
*		rfsr	The value read from the Rx FIFO Size Register
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_fs(__u32 frr, __u32 fsr, __u32 tfsr, __u32 rfsr)
{
	int			errs	= 0;
	const char*	str;

	label("FIFO Sizes");

	if ((frr & 0x80000000) == 0)
	{
		printf("(not specified by firmware)\n");
	}
	else
	{
		printf("\n");
		label_level_inc();

		label("Tx FIFO Size");
		str	= label_format_k(tfsr, 1024);
		printf("%s Samples\n", str);

		label("Rx FIFO Size");
		str	= label_format_k(rfsr, 1024);
		printf("%s Samples\n", str);

		label_level_dec();
	}

	return(errs);
}



/*****************************************************************************
*
*	Function:	_id_board_fsr
*
*	Purpose:
*
*		Identify the board from the Feature Set Register.
*
*	Arguments:
*
*		frr		The value read from the Firmware Revision Register.
*
*		fsr		The value read from the Feature Set Register.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_fsr(__u32 frr, __u32 fsr)
{
	int		errs	= 0;
	__u32	fld;

	label("Feature Set Register");

	if ((frr & 0x80000000) == 0)
	{
		printf("(absent)\n");
	}
	else
	{
		printf("0x%08lX\n", (unsigned long) fsr);
		label_level_inc();

		label("FIFO Size Regs");
		fld	= fsr & 0x1;
		printf("0x-------%01lX %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("FIFO Words Regs");
		fld	= fsr & 0x2;
		printf("0x-------%01lX %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("IRQ Config. Regs");
		fld	= fsr & 0x4;
		printf("0x-------%01lX %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("GPIO 0-5");
		fld	= fsr & 0x8;
		printf("0x-------%01lX %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("DMDMA on Ch 1");
		fld	= (fsr >> 4) & 0x1;
		printf("0x------%01lX- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("Tx/Rx Over/Underrun");
		fld	= (fsr >> 4) & 0x2;
		printf("0x------%01lX- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("GPIO 6");
		fld	= (fsr >> 4) & 0x4;
		printf("0x------%01lX- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("Tx Auto Stop Disable");
		fld	= (fsr >> 4) & 0x4;
		printf("0x------%01lX- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("Single Cycle Disable");
		fld	= (fsr >> 4) & 0x8;
		printf("0x------%01lX- %s\n", (long) fld, fld ? "(Yes)" : "(No)");

		label("Reserved");
		fld	= fsr >> 8;
		printf("0x%06lX-- %s\n", (long) fld, fld ? "(UNRECOGNIZED)" : "");

		label_level_dec();
	}

	return(errs);
}



/*****************************************************************************
*
*	Function:	_id_board_pci
*
*	Purpose:
*
*		Identify the board from the PCI registers.
*
*	Arguments:
*
*		idr		The PCI Id Register value.
*
*		svid	The PCI Subsystem Vendor Id Register value.
*
*		sid		The PCI Subsystem Id Register value.
*
*	Returned:
*
*		>= 0	The number of errors seen.
*
*****************************************************************************/

static int _id_board_pci(__u32 idr, __u32 svid, __u32 sid)
{
	__u32	fld;
	int		errs	= 0;

	label("Vendor ID");
	fld	= idr & 0xFFFF;
	printf("0x%04lX     ", (long) fld);

	if (fld == HPDI32_VENDOR_ID)
	{
		printf("(PLX)\n");
	}
	else
	{
		errs++;
		printf("(UNKNOWN) FAIL <---\n");
	}

	label("Product ID");
	fld	= idr >> 16;
	printf("0x%04lX     ", (long) fld);

	if (fld == HPDI32_DEVICE_ID_32)
	{
		printf("(PCI9080)\n");
	}
	else if (fld == HPDI32_DEVICE_ID_64)
	{
		printf("(PCI9656)\n");
	}
	else
	{
		errs++;
		printf("(UNKNOWN) FAIL <---\n");
	}

	label("Sub Vendor ID");
	printf("0x%04lX     ", (long) svid);

	if (svid == HPDI32_SUBVENDOR_ID)
	{
		printf("(PLX Assigned Subsystem ID)\n");
	}
	else
	{
		errs++;
		printf("(UNKNOWN) FAIL <---\n");
	}

	label("Subsystem ID");
	printf("0x%04lX     ", (long) sid);

	if (sid == HPDI32_SUBDEVICE_ID_32)
	{
		printf("(HPDI32 (32-bit))\n");
	}
	else if (sid == HPDI32_SUBDEVICE_ID_64)
	{
		printf("(HPDI32 (64-bit))\n");
	}
	else
	{
		errs++;
		printf("(UNKNOWN) FAIL <---\n");
	}

	return(errs);
}



/*****************************************************************************
*
*	Function:	id_board
*
*	Purpose:
*
*		Identify the board.
*
*	Arguments:
*
*		fd		The handle to use to access the board.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
*****************************************************************************/

int id_board(int fd)
{
	__u32	did;
	int		errs	= 0;
	__u32	frr;
	__u32	fsr;
	__u32	idr;
	__u32	rfsr;
	__u32	sid;
	__u32	svid;
	__u32	tfsr;

	label("Board Identification");
	errs	+= hpdi32_reg_read(fd, HPDI32_PCI_IDR, &idr, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_PCI_SVID, &svid, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_PCI_SID, &sid, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_GSC_FRR, &frr, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_GSC_FSR, &fsr, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_GSC_RFSR, &rfsr, 0);
	errs	+= hpdi32_reg_read(fd, HPDI32_GSC_TFSR, &tfsr, 0);
	errs	= errs ? 1 : 0;

	if (errs == 0)
	{
		did	= idr >> 16;
		printf("\n");
		label_level_inc();
		errs	+= _id_board_pci(idr, svid, sid);
		errs	+= _id_board_frr(did, frr);
		errs	+= _id_board_fsr(frr, fsr);
		errs	+= _id_board_fs(frr, fsr, tfsr, rfsr);
		label_level_dec();
	}

	return(errs);
}



/*****************************************************************************
*
*	Function:	id_driver
*
*	Purpose:
*
*		Identify the driver.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
*****************************************************************************/

int id_driver(int fd)
{
	int						errs;
	hpdi32_driver_info_t	info;

	label("Driver");
	memset(&info, 0, sizeof(info));
	errs	= hpdi32_driver_info(fd, &info, 1);

	if (errs == 0)
	{
		printf("\n");
		label_level_inc();
		label("Version");
		printf("%s\n", info.version);
		label("Built");
		printf("%s\n", info.built);
		label_level_dec();
	}

	return(errs);
}



/*****************************************************************************
*
*	Function:	id_host
*
*	Purpose:
*
*		Identify the operating system.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

void id_host(void)
{
	char			buf[1024];
	int				end;
	FILE*			file;
	int				i;
	int				index;
	char*			psz;
	char			tmp[1024];
	struct utsname	uts;

	memset(&uts, 0, sizeof(uts));
	uname(&uts);
	label("Host");
	printf("\n");
	label_level_inc();

	label("Operating System");
	_print_str(uts.sysname, "\n");
	label("Kernel");
	_print_str(uts.release, "\n");
	label("Processor");
	_print_str(uts.machine, "\n");
	label_level_inc();

	file	= fopen("/proc/cpuinfo", "r");

	if (file)
	{
		for (;;)
		{
			end	= feof(file);

			if (end)
				break;

			fgets(buf, sizeof(buf), file);
			i	= sscanf(buf, "processor : %d", &index);

			if (i == 1)
			{
				label_level_dec();
				sprintf(buf, "Processor #%d", index);
				label(buf);
				printf("\n");
				label_level_inc();
				continue;
			}

			i	= sscanf(buf, "model name : %s", tmp);

			if (i == 1)
			{
				label("CPU Model");
				psz	= strchr(buf, ':') + 2;
				_print_str(psz, "\n");
				continue;
			}

			i	= sscanf(buf, "cpu MHz : %s", tmp);

			if (i == 1)
			{
				label("CPU Speed");
				_print_str(tmp, " MHz\n");
				continue;
			}
		}

		fclose(file);
	}

	label_level_dec();
	label_level_dec();
}



